"""
trionyx.views
~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import csv
import io
import json
import logging
from collections import OrderedDict

from django.apps import apps
from django.views.generic import (
    View,
    TemplateView,
    DetailView,
    UpdateView as
    DjangoUpdateView,
    CreateView as DjangoCreateView,
    DeleteView as DjangoDeleteView
)
from django.core.exceptions import PermissionDenied
from django.http import Http404, StreamingHttpResponse
from django.urls import reverse
from django.core.paginator import Paginator
from django.contrib import messages
from watson import search as watson
from django.template.loader import render_to_string
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from trionyx.views.mixins import ModelClassMixin, SessionValueMixin, ModelPermissionMixin
from trionyx.forms.helper import FormHelper
from trionyx.models import filter_queryset_with_user_filters
from .ajax import JsendView

logger = logging.getLogger(__name__)


# =============================================================================
# List view
# =============================================================================
class ListView(ModelPermissionMixin, TemplateView, ModelClassMixin):
    """List view for showing model"""

    permission_type = 'view'
    template_name = "trionyx/core/model_list.html"

    model = None
    """Model class default will try get model based on url kwargs app, model"""

    title = None
    """Title of page default is model verbose_name_plural"""

    ajax_url = None
    """Ajax url used to get model list data"""

    download_url = None
    """Download url used to get model list data"""

    def get_title(self):
        """Get page title"""
        if self.title:
            return self.title
        return self.get_model_class()._meta.verbose_name_plural

    def get_ajax_url(self):
        """Get ajax url"""
        if self.ajax_url:
            return self.ajax_url
        return reverse('trionyx:model-list-ajax', kwargs=self.kwargs)

    def get_download_url(self):
        """Get ajax url"""
        if self.download_url:
            return self.download_url
        return reverse('trionyx:model-list-download', kwargs=self.kwargs)

    def get_create_url(self):
        """Get create url"""
        return reverse('trionyx:model-create', kwargs=self.kwargs)

    def get_mass_delete_url(self):
        """Get mass delete url"""
        return reverse('trionyx:model-mass-delete', kwargs=self.kwargs)

    def get_mass_update_url(self):
        """Get mass delete url"""
        return reverse('trionyx:model-mass-update', kwargs=self.kwargs)

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)
        context.update({
            'title': self.get_title(),
            'content_type_id': ContentType.objects.get_for_model(self.get_model_class()).id,
            'model_config': self.get_model_config(),
            'ajax_url': self.get_ajax_url(),
            'download_url': self.get_download_url(),
            'create_url': self.get_create_url(),
            'header_buttons': self.get_model_config().get_header_buttons(context={
                'page': 'list',
            }),
            'mass_delete_url': self.get_mass_delete_url(),
            'mass_update_url': self.get_mass_update_url(),
            'create_permission': self.get_model_config().has_permission('add', user=self.request.user),
            'change_permission': self.get_model_config().has_permission('change', user=self.request.user),
            'delete_permission': self.get_model_config().has_permission('delete', user=self.request.user),
        })
        return context


class ModelListMixin(ModelClassMixin, SessionValueMixin):
    """Mixin for handling model list"""

    def get_page(self, paginator):
        """Get current page or page in session"""
        page = int(self.get_and_save_value('page', 1))
        if page < 1:
            return self.save_value('page', 1)

        if page > paginator.num_pages:
            return self.save_value('page', paginator.num_pages)
        return page

    def get_page_size(self):
        """Get current page size or page size in session"""
        return self.get_and_save_value('page_size', '10')

    def get_sort(self):
        """Get current sort or sort in session"""
        return self.get_and_save_value('sort', self.get_model_config().list_default_sort)

    def get_search(self):
        """Get current search or search from session, reset page if search is changed"""
        old_search = self.get_session_value('search', '')
        search = self.get_and_save_value('search', '')
        if old_search != search:
            self.page = 1
            self.get_session_value('page', self.page)
        return search

    def get_filters(self):
        """Get all active filters"""
        try:
            return json.loads(self.get_and_save_value('filters', '[]'))
        except json.JSONDecodeError:
            return []

    def get_all_fields(self):
        """Get all available fields"""
        return {
            name: {
                'name': name,
                'label': field['label'],
                'type': field['type'],
                'choices': field['choices'],
            }
            for name, field in self.get_model_config().get_list_fields().items()
        }

    def get_current_fields(self):
        """Get current list to be used"""
        if hasattr(self, 'current_fields') and self.current_fields:
            return self.current_fields

        field_attribute = 'list_{}_{}_fields'.format(self.kwargs.get('app'), self.kwargs.get('model'))
        current_fields = self.request.user.get_attribute(field_attribute, [])
        request_fields = self.request.POST.get('selected_fields', None)

        if request_fields and ','.join(current_fields) != request_fields:
            # TODO validate fields
            current_fields = list(filter(lambda f: f != 'NR', request_fields.split(',')))
            self.request.user.set_attribute(field_attribute, current_fields)
        elif request_fields:
            current_fields = request_fields.split(',')

        if not current_fields:
            config = self.get_model_config()
            current_fields = config.list_default_fields if config.list_default_fields else ['id']

        # TODO check if all fields are still valid, filter out invalid and save
        self.current_fields = current_fields
        return current_fields

    def get_paginator(self):
        """Get paginator"""
        return Paginator(self.get_queryset(), self.get_page_size())

    def get_queryset(self):
        """Get qeuryset for model"""
        query = self.search_queryset()
        query = filter_queryset_with_user_filters(query, self.get_filters(), self.request)

        fields = self.get_all_fields()
        select_related = self.get_model_config().list_select_related if self.get_model_config().list_select_related else []
        for field in self.get_current_fields():
            field_parts = field.split('__')
            if len(field_parts) > 1:
                select_related.append('__'.join(field_parts[:-1]))
            elif fields[field]['type'] == 'related':
                select_related.append(field)

        if select_related:
            query = query.select_related(*set(select_related))

        return query.order_by(self.get_sort())

    def search_queryset(self):
        """Get search query set"""
        queryset = self.get_model_class().objects.get_queryset()
        return watson.filter(queryset, self.get_search(), ranking=False)


class ListJsendView(ModelPermissionMixin, JsendView, ModelListMixin):
    """Ajax list view"""

    permission_type = 'view'

    def __init__(self, *args, **kwargs):
        """Init ListJsendView"""
        super().__init__(*args, **kwargs)
        self.page = None
        self.page_size = None
        self.sort = None
        self.fields = None

    def handle_request(self, request, *args, **kwargs):
        """Give back list items + config"""
        paginator = self.get_paginator()
        # Call search first, it will reset page if search is changed
        search = self.get_search()
        page = self.get_page(paginator)
        items = self.get_items(paginator, page)
        return {
            'search': search,
            'filters': self.get_filters(),
            'page': page,
            'page_size': self.get_page_size(),
            'num_pages': paginator.num_pages,
            'count': paginator.count,
            'sort': self.get_sort(),
            'current_fields': self.get_current_fields(),
            'fields': self.get_all_fields(),

            'items': items,
        }

    def get_items(self, paginator, current_page):
        """Get list items for current page"""
        fields = self.get_model_config().get_list_fields()

        page = paginator.page(current_page)

        items = []
        for item in page:
            items.append({
                'id': item.id,
                'url': self.get_model_config().get_absolute_url(item),
                'row_data': [
                    fields[field]['renderer'](item, field)
                    for field in self.get_current_fields()
                ]
            })
        return items


class ListExportView(ModelPermissionMixin, View, ModelListMixin):
    """View for downloading an export of a list view"""

    permission_type = 'view'

    def post(self, request, app, model, **kwargs):
        """Handle post request"""
        return self.csv_response()

    def items(self):
        """Get all list items"""
        query = self.get_queryset()
        fields = self.get_model_config().get_list_fields()

        for item in query.iterator():
            row = OrderedDict()
            for field_name in self.get_current_fields():
                field = fields.get(field_name)
                if not field_name:
                    row[field_name] = ''

                if hasattr(item, field['field']):
                    row[field_name] = getattr(item, field['field'])
                else:
                    row[field_name] = ''  # TODO Maybe render field ans strip html?
            yield row

    def csv_response(self):
        """Get csv response"""
        def stream():
            """Create data stream generator"""
            stream_file = io.StringIO()
            csvwriter = csv.writer(stream_file, delimiter=',', quotechar='"')

            csvwriter.writerow(self.get_current_fields())

            for index, item in enumerate(self.items()):
                csvwriter.writerow([value for index, value in item.items()])
                stream_file.seek(0)
                data = stream_file.read()
                stream_file.seek(0)
                stream_file.truncate()
                yield data

        response = StreamingHttpResponse(stream(), content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={}.csv".format(self.get_model_config().model_name.lower())
        return response


class ListChoicesJsendView(ModelPermissionMixin, JsendView, ModelListMixin):
    """View for getting choices list for related field"""

    permission_type = 'view'

    def handle_request(self, request, *args, **kwargs):
        """Build choices list for related field"""
        try:
            ModelClass = self.get_model_class()
            fields = request.GET.get('field').split('__')
            RelatedClass = getattr(ModelClass, fields[0]).field.related_model

            for field in fields[1:]:
                RelatedClass = getattr(RelatedClass, field).field.related_model
        except Exception:
            return []
        return list(RelatedClass.objects.values_list('id', 'verbose_name'))


# =============================================================================
# Detail tab views
# =============================================================================
class DetailTabView(ModelPermissionMixin, DetailView, ModelClassMixin):
    """Detail tab view, shows model details in tab view"""

    permission_type = 'view'

    template_name = "trionyx/core/model_view.html"
    """Template name for rendering the view, default is 'trionyx/core/detail_tab_view.html'

        .. note::
            Only overide this if you want to customize the layout of the page.
    """

    model_alias: str = ''
    """Model identifier, default is '<model app name>.<model name>'."""

    title: str = ''
    """Page title if not set object __str__ is used"""

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                ModelClass = apps.get_model(self.kwargs.get('app'), self.kwargs.get('model'))
                return ModelClass._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)
        tabs = self.get_active_tabs()
        context.update({
            'model_config': self.get_model_config(),
            'page_detail_tabs': tabs,
            'active_tab': tabs[0].code if tabs else '',
            'app_label': self.get_app_label(),
            'model_name': self.get_model_name(),
            'model_alias': self.get_model_alias(),
            'model_verbose_name': self.object._meta.verbose_name.title(),
            'back_url': self.get_back_url(),
            'edit_url': self.get_edit_url(),
            'delete_url': self.get_delete_url(),
            'title': self.title,
            'change_permission': self.get_model_config().has_permission('change', self.object, user=self.request.user),
            'delete_permission': self.get_model_config().has_permission('delete', self.object, user=self.request.user),
        })
        return context

    def get_back_url(self):
        """Get back url"""
        return reverse('trionyx:model-list', kwargs={
            'app': self.get_app_label(),
            'model': self.get_model_name(),
        })

    def get_delete_url(self):
        """Get model object delete url"""
        return reverse('trionyx:model-delete', kwargs={
            'app': self.get_app_label(),
            'model': self.get_model_name(),
            'pk': self.object.id
        })

    def get_edit_url(self):
        """Get model object edit url"""
        return reverse('trionyx:model-edit', kwargs={
            'app': self.get_app_label(),
            'model': self.get_model_name(),
            'pk': self.object.id
        })

    def get_model_alias(self):
        """Get model alias"""
        if self.model_alias:
            return self.model_alias
        return '{}.{}'.format(self.get_app_label(), self.get_model_name())

    def get_app_label(self):
        """Get model app label"""
        return self.object._meta.app_label

    def get_model_name(self):
        """Get model name"""
        return self.object._meta.model_name

    def get_active_tabs(self):
        """Get all active tabs"""
        from trionyx.views import tabs
        return list(tabs.get_tabs(self.get_model_alias(), self.object))

    def dispatch(self, request, *args, **kwargs):
        """Validate if user can use view"""
        if False:  # TODO do permission check based on Model
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)


class DetailTabJsendView(ModelPermissionMixin, JsendView, ModelClassMixin):
    """View for getting tab view with ajax"""

    permission_type = 'view'

    def handle_request(self, request, app, model, pk):
        """Render and return tab"""
        from trionyx.views import tabs

        tab_code = request.GET.get('tab')
        model_alias = request.GET.get('model_alias')
        model_alias = model_alias if model_alias else '{}.{}'.format(app, model)

        # TODO permission check

        item = tabs.get_tab(model_alias, self.get_object(), tab_code)

        return {
            'header_buttons': render_to_string('trionyx/base/model_header_buttons.html', {
                'header_buttons': self.get_model_config().get_header_buttons(self.get_object(), {
                    'page': 'view',
                    'model_alias': model_alias,
                    'tab': tab_code,
                }),
            }, request=request),
            'content': item.get_layout(self.get_object()).render(request),
        }

    def get_object(self):
        """Get object"""
        if not hasattr(self, 'object'):
            ModelClass = self.get_model_class()
            self.object = ModelClass.objects.get(id=self.kwargs.get('pk'))
        return self.object


class LayoutView(DetailTabView):
    """Display layout for model"""

    permission_type = 'view'

    template_name = "trionyx/core/layout_view.html"

    def get_context_data(self, **kwargs):
        """Add layout to context"""
        context = super().get_context_data(**kwargs)
        from trionyx.views import layouts
        try:
            context['layout'] = layouts.get_layout(self.kwargs.get('code'), self.object).render(self.request)
        except Exception:
            raise Http404()
        return context


class LayoutUpdateView(JsendView, ModelClassMixin):
    """Update whole or section of an layout"""

    def handle_request(self, request, pk, code, **kwargs):
        """Render layout or given component for layout"""
        from trionyx.views import layouts
        layout = layouts.get_layout(
            code,
            self.get_model_class().objects.get(id=pk),
            request.GET.get('layout_id')
        )

        component_id = request.GET.get('component')
        if component_id:
            comp, _ = layout.find_component_by_id(component_id)
            return comp.render({}, request)
        else:
            return ''.join(comp.render({}, request) for comp in layout.components)


# =============================================================================
# Update/Create/Delete view
# =============================================================================
class UpdateView(ModelPermissionMixin, DjangoUpdateView, ModelClassMixin):
    """Update view that renders view with crispy-forms"""

    permission_type = 'change'

    template_name = 'trionyx/core/model_update.html'

    title: str = ''
    """Title of page, default is: Update <object.__str__>"""

    submit_value: str = ''
    """Value of the form submit button"""

    cancel_url: str = ''
    """Url code for cancel button, when not set object.get_absolute_url is used"""

    @property
    def success_url(self):
        """Return success url"""
        return self.get_model_config().get_absolute_url(self.object)

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                return self.get_model_class()._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_form_class(self):
        """Get form class for model"""
        if self.form_class:
            return self.form_class
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_edit_form(self.get_model_class())

    def get_form(self, form_class=None):
        """Get form for model"""
        if not hasattr(self, 'form'):
            self.form = super().get_form(form_class)

        if not getattr(self.form, 'helper', None):
            self.form.helper = FormHelper(self.form)
            self.form.helper.form_tag = False
        else:
            self.form.helper.form_tag = False
        return self.form

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)

        context.update({
            'title': self.get_form().get_title() if hasattr(self.get_form(), 'get_title') else self.title,
            'submit_value': (self.get_form().get_submit_label()
                             if hasattr(self.get_form(), 'get_submit_label')
                             else self.submit_value),
            'header_buttons': list(
                self.get_model_config().get_header_buttons(self.object, {'page': 'edit'})
            ),
            'cancel_url': self.cancel_url,
            'object_url': self.get_model_config().get_absolute_url(self.object),
        })

        return context

    def form_valid(self, form):
        """Add success message"""
        response = super().form_valid(form)
        messages.success(self.request, _("Successfully saved ({})").format(self.object))
        return response

    def form_invalid(self, form):
        """Form invalid"""
        logger.debug(json.dumps(form.errors))
        return super().form_invalid(form)


class CreateView(ModelPermissionMixin, DjangoCreateView, ModelClassMixin):
    """Create view that renders view with crispy-forms"""

    permission_type = 'add'

    template_name = 'trionyx/core/model_create.html'

    title = None
    """Title of page, default is: Update <object.__str__>"""

    submit_value = None
    """Value of the form submit button"""

    cancel_url = None
    """Url code for cancel button, when not set model list view is used"""

    @property
    def success_url(self):
        """Return success url"""
        return self.get_model_config().get_absolute_url(self.object)

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                return self.get_model_class()._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_form_class(self):
        """Get form class for model"""
        if self.form_class:
            return self.form_class
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_create_form(self.get_model_class())

    def get_form(self, form_class=None):
        """Get form for model"""
        if not hasattr(self, 'form'):
            self.form = super().get_form(form_class)

        if not getattr(self.form, 'helper', None):
            self.form.helper = FormHelper(self.form)
            self.form.helper.form_tag = False
        else:
            self.form.helper.form_tag = False
        return self.form

    def get_cancel_url(self):
        """Get cancel url"""
        if self.cancel_url:
            return self.cancel_url

        ModelClass = self.get_model_class()
        return reverse('trionyx:model-list', kwargs={
            'app': ModelClass._meta.app_label,
            'model': ModelClass._meta.model_name,
        })

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)

        context.update({
            'title': self.get_form().get_title() if hasattr(self.get_form(), 'get_title') else self.title,
            'submit_value': (self.get_form().get_submit_label()
                             if hasattr(self.get_form(), 'get_submit_label')
                             else self.submit_value),
            'cancel_url': self.get_cancel_url(),
            'model_verbose_name': self.get_model_class()._meta.verbose_name,
        })

        return context

    def form_valid(self, form):
        """Add success message"""
        response = super().form_valid(form)
        messages.success(self.request, _("Successfully created ({})").format(self.object))
        return response

    def form_invalid(self, form):
        """Form invalid"""
        logger.debug(json.dumps(form.errors))
        return super().form_invalid(form)


class DeleteView(ModelPermissionMixin, DjangoDeleteView, ModelClassMixin):
    """Delete view"""

    permission_type = 'delete'

    template_name = 'trionyx/core/model_delete.html'

    title = None
    """Title of page, default is: Update <object.__str__>"""

    submit_value = None
    """Value of the form submit button"""

    cancel_url = None
    """Url code for cancel button, when not set model list view is used"""

    def get_queryset(self):
        """Get queryset based on url params(<app>, <mode>) if model is not set on class"""
        if self.queryset is None and not self.model:
            try:
                return self.get_model_class()._default_manager.all()
            except LookupError:
                raise Http404()
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        """Add context data to view"""
        context = super().get_context_data(**kwargs)
        context.update({
            'title': self.title,
            'submit_value': self.submit_value,
            'cancel_url': self.cancel_url,
            'object_url': self.get_model_config().get_absolute_url(self.object)
        })
        return context

    def get_success_url(self):
        """Get success url"""
        messages.success(self.request, _("Successfully deleted ({})").format(self.object))
        if self.success_url:
            return reverse(self.success_url)

        if 'app' in self.kwargs and 'model' in self.kwargs:
            return reverse('trionyx:model-list', kwargs={
                'app': self.kwargs.get('app'),
                'model': self.kwargs.get('model'),
            })

        return '/'



"""Base Project settings"""
from trionyx.settings import *  # noqa F403 F401


TX_APP_NAME = '[[app_name]]'
TX_LOGO_NAME_START = '[[logo_name_start]]'
TX_LOGO_NAME_END = '[[logo_name_end]]'
TX_LOGO_NAME_SMALL_START = '[[logo_name_small_start]]'
TX_LOGO_NAME_SMALL_END = '[[logo_name_small_end]]'

TX_COMPANY_NAME = '[[app_name]]'

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

STATIC_ROOT = os.path.join(BASE_DIR, 'static')
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

INSTALLED_APPS += [ # noqa F405
    # Extra apps
]

ROOT_URLCONF = 'config.urls'

LOCALE_PATHS = (
    os.path.join(BASE_DIR, 'locale'),
)

WATSON_POSTGRES_SEARCH_CONFIG = get_watson_search_config(LANGUAGE_CODE)

# Celery beat schedule
from trionyx.utils import create_celerybeat_schedule  # noqa E402
CELERY_BEAT_SCHEDULE = create_celerybeat_schedule(INSTALLED_APPS)



# Generated by Django 2.2.6 on 2019-11-17 10:30

from django.db import migrations, models
import jsonfield.encoder
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('trionyx', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SystemVariable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=128, unique=True)),
                ('value', jsonfield.fields.JSONField(dump_kwargs={'cls': jsonfield.encoder.JSONEncoder, 'separators': (',', ':')}, load_kwargs={})),
            ],
        ),
    ]



"""
trionyx.trionyx
~~~~~~~~~~~~~~~

Core app for Trionyx

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import threading


LOCAL_DATA = threading.local()
"""Data storage for current request thread"""

default_app_config = 'trionyx.trionyx.apps.Config'



"""
trionyx.views.mixins
~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
from typing import Optional, Type, Any

from django.apps import apps
from django.db.models import Model
from django.http.request import HttpRequest
from django.core.exceptions import PermissionDenied
from django.http import Http404

from trionyx.config import models_config, ModelConfig


class ModelClassMixin:
    """Mixen for getting model class"""

    def get_model_class(self) -> Optional[Type[Model]]:
        """Get model class"""
        kwargs = getattr(self, 'kwargs', {})

        if getattr(self, 'model', None):
            return getattr(self, 'model')
        elif getattr(self, 'object', None):
            return getattr(self, 'object').__class__
        elif 'app' in kwargs and 'model' in kwargs:
            try:
                return apps.get_model(kwargs.get('app'), kwargs.get('model'))  # type: ignore
            except LookupError:
                raise Http404()
        elif hasattr(self, 'get_queryset'):
            return self.get_queryset().model  # type: ignore
        else:
            raise Http404()

    def get_model_config(self) -> ModelConfig:
        """Get Trionyx model config"""
        ModelClass = self.get_model_class()
        if not ModelClass:
            raise Exception('Could not get model class')

        if not hasattr(self, '__config'):
            setattr(self, '__config', models_config.get_config(ModelClass))
        return getattr(self, '__config', None)


class ModelPermissionMixin:
    """Check Model permission"""

    permission_type: Optional[str] = None

    def get_object(self, *args, **kwargs):
        """Override to prevent multiple lookups"""
        if not getattr(self, 'object', False) and hasattr(super(), 'get_object'):
            self.object = getattr(super(), 'get_object', lambda *args, **kwargs: None)(*args, **kwargs)
        return getattr(self, 'object', None)

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        """Validate if user can use view"""
        if self.permission_type in ['view', 'add', 'change', 'delete'] and hasattr(self, 'get_model_config'):
            try:
                obj = self.get_object()
            except AttributeError:
                obj = None
            if not self.get_model_config().has_permission(self.permission_type, obj, request.user):  # type: ignore
                raise PermissionDenied()
        elif self.permission and not request.user.has_perm(self.permission):
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)  # type: ignore

    @property
    def permission(self) -> Optional[str]:
        """Permission for view"""
        if self.permission_type and hasattr(self, 'get_model_config'):
            config = self.get_model_config()  # type: ignore

            return '{app_label}.{type}_{model_name}'.format(
                app_label=config.app_label,
                type=self.permission_type,
                model_name=config.model_name,
            ).lower()
        return None


class SessionValueMixin:
    """Mixin for handling session values"""

    def get_and_save_value(self, name: str, default: Any = None) -> Any:
        """Get value from request/session and save value to session"""
        request: HttpRequest = getattr(self, 'request')

        if getattr(self, name, None):
            return getattr(self, name)

        value = self.get_session_value(name, default)
        value = request.POST.get(name, value)
        self.save_value(name, value)
        return value

    def get_session_value(self, name: str, default: Any = None) -> Any:
        """Get value from session"""
        kwargs = getattr(self, 'kwargs', {})
        request: HttpRequest = getattr(self, 'request')
        session_name = 'list_{}_{}_{}'.format(kwargs.get('app'), kwargs.get('model'), name)
        return request.session.get(session_name, default)

    def save_value(self, name: str, value: Any) -> Any:
        """Save value to session"""
        kwargs = getattr(self, 'kwargs', {})
        request: HttpRequest = getattr(self, 'request')
        session_name = 'list_{}_{}_{}'.format(kwargs.get('app'), kwargs.get('model'), name)
        request.session[session_name] = value
        setattr(self, name, value)
        return value



from datetime import datetime

from django.test import TestCase
from trionyx import layout as l  # noqa E741

from trionyx.trionyx.models import User


class ModelsTest(TestCase):
    def setUp(self):
        self.user = User.objects.create(email="info@trionyx.com")
        self.layout = l.Layout(
            l.Row(
                l.Column6(
                    l.Panel(
                        'Left panel',
                        id='left-panel'
                    )
                ),
                l.Column6(
                    l.Panel(
                        'Right panel',
                        id='right-panel'
                    )
                )
            )
        )

    def test_layout_path(self):
        panel, parent = self.layout.find_component_by_path('row.column6[1].panel')
        self.assertEqual(panel.title, 'Right panel')

    def test_layout_id(self):
        panel, parent = self.layout.find_component_by_id('right-panel')
        self.assertEqual(panel.title, 'Right panel')

    def test_remove_component(self):
        self.layout.delete_component(path='row.column6.panel')
        self.layout.delete_component(id='right-panel')
        self.assertInHTML("""
        <div class="row">
            <div class="col-md-6"></div>
            <div class="col-md-6"></div>
        </div>
        """, self.layout.render())

    def test_remove_top_component(self):
        self.layout.delete_component(path='row')
        self.assertInHTML("""<div id="{}"></div>""".format(self.layout.id), self.layout.render())

    def test_remove_invalid_component(self):
        with self.assertRaises(Exception):
            self.layout.delete_component(path='some random path')

    def test_remove_no_params_component(self):
        with self.assertRaises(Exception):
            self.layout.delete_component()

    def test_add_component(self):
        self.layout.add_component(l.Img(), path='row.column6[1].panel', append=True)
        self.layout.add_component(l.Img(), id='left-panel', append=True)

        self.assertInHTML("""
        <div class="row">
            <div class="col-md-6">
                <div id="component-left-panel">
                    <div class="panel panel-default">
                        <div class="panel-heading">Left panel</div>
                        <div class="panel-collapse">
                            <img width="100%" />
                        </div>
                    </div>
                </div>
            </div>
            <div class="col-md-6">
                <div id="component-right-panel">
                    <div class="panel panel-default">
                        <div class="panel-heading">Right panel</div>
                        <div class="panel-collapse">
                            <img width="100%" />
                        </div>
                    </div>
                </div>
            </div>
        </div>
        """, self.layout.render())

    def test_input(self):
        self.assertHTMLEqual(l.Input(name='test').render({}), """
        <div class="form-group no-margin ">
            <input type="text" class="form-control" name="test" />
        </div>
        """)

    def test_badge(self):
        badge = l.Badge(html='test')
        badge.set_object(self.user)
        self.assertHTMLEqual(
            badge.render({}),
            f"""<span class="badge bg-theme">test</span>""")

    def test_alert(self):
        self.assertHTMLEqual(l.Alert('trionyx', no_margin=True).render({}), """
        <div class="alert alert-success no-margin">
            trionyx
        </div>
        """)

    def test_button(self):
        button = l.Button('Trionyx')
        button.set_object(self.user)
        self.assertHTMLEqual(button.render({}), """
        <button class="btn btn-flat bg-theme" onclick="window.location.href='/model/trionyx/user/{id}/'; return false;">
            Trionyx
        </button>
        """.format(id=self.user.id))

    def test_button_dialog(self):
        button = l.Button('Trionyx', dialog=True, dialog_reload_tab='general')
        button.set_object(self.user)
        self.assertHTMLEqual(button.render({}), """
        <button class="btn btn-flat bg-theme" onClick="openDialog('/model/trionyx/user/{id}/', {{ callback:function(data, dialog){{
                if (data.success) {{
                    dialog.close();
                    trionyx_reload_tab('general');
                }}
            }} }}); return false;">
            Trionyx
        </button>
        """.format(id=self.user.id))

    def test_ordered_list(self):
        unordered = l.OrderedList(
            {
                'value': 'Item 1',
            },
            {
                'value': 'Item 2',
            }
        )
        unordered.set_object({})
        self.assertHTMLEqual(unordered.render({}), """
        <ol>
            <li>
                Item 1
            </li><li>
                Item 2
            </li>
        </ol>
        """)

    def test_ordered_list_objects(self):
        unordered = l.OrderedList(
            '_index_',
            objects=[['Item 1'], ['Item 2']]
        )
        unordered.set_object({})
        self.assertHTMLEqual(unordered.render({}), """
        <ol>
            <li>
                Item 1
            </li><li>
                Item 2
            </li>
        </ol>
        """)

    def test_progressBar(self):
        progressbar = l.ProgressBar(value=10)
        progressbar.set_object({})
        self.assertHTMLEqual(progressbar.render({}), """
        <div class=" progress progress-md">
            <div aria-valuemax="100" aria-valuemin="0" aria-valuenow="10"
            class=" progress-bar progress-bar-theme" role="progressbar" style="width: 10%">
                <span>
                    10%
                </span>
            </div>
        </div>
        """)

    def test_table_description(self):
        self.assertHTMLEqual(l.TableDescription({
            'label': 'Trionyx',
            'value': 'Test',
        }).render({}), """
        <table class="table table-condensed">
            <tbody>
                <tr>
                    <th style="width: 150px;">
                        Trionyx:
                    </th>
                    <td class="description-value">
                        Test
                    </td>
                </tr>
            </tbody>
        </table>
        """)

    def test_table(self):
        html = l.Table(
            [['column1', 'column2', 'column3']],
            {
                'label': 'Column 1',
            },
            {
                'label': 'Column 2',
            },
            {
                'label': 'Column 3',
            },
            footer=[
                [
                    ['Footer', '123'],
                ],
                {
                    'colspan': 2,
                },
                {
                    'class': 'text-right',
                }
            ]
        ).render({})
        self.assertHTMLEqual(html, """
        <div class="table-responsive">
            <table class="table table-condensed">
                <thead>
                    <tr>
                        <th>Column 1</th>
                        <th>Column 2</th>
                        <th>Column 3</th>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>column1</td>
                        <td>column2</td>
                        <td>column3</td>
                    </tr>
                </tbody>
                <tfoot>
                    <tr>
                        <td colspan="2">Footer</td>
                        <td class="text-right">123</td>
                    </tr>
                </tfoot>
            </table>
        </div>
        """)

    def test_line_chart(self):
        chart = l.LineChart(
            [
                [datetime.fromtimestamp(1576951711), 1],
                [datetime.fromtimestamp(1576941711), 2],
                [datetime.fromtimestamp(1576931711), 3]
            ],
            'x',
            'y'
        )
        chart.set_object({})

        self.assertEqual(chart.chart_data['labels'], [1576951711000, 1576941711000, 1576931711000])
        self.assertEqual(chart.chart_data['datasets'][0]['data'], [
            {'x': 1576951711000, 'y': 1, 'label': '1'},
            {'x': 1576941711000, 'y': 2, 'label': '2'},
            {'x': 1576931711000, 'y': 3, 'label': '3'}
        ])

    def test_pie_chart(self):
        chart = l.PieChart(
            [['Python3', 60], ['Javascript', 30], ['Sql', 10]],
            'name',
            'value',
        )
        chart.set_object({})

        self.assertEqual(chart.chart_data['labels'], ['Python3', 'Javascript', 'Sql'])
        self.assertEqual(chart.chart_data['datasets'][0]['data'], [60, 30, 10])



from django.test import TestCase

from trionyx.tasks import BaseTask
from trionyx.trionyx.models import Task, User


class TestTask(BaseTask):
    def run(self):
        self.set_progress(50)
        self.add_output('Halfway')
        if not self.get_object():
            raise Exception('no object')

        return 'some results'


class TasksTest(TestCase):

    def setUp(self):
        self.user = User.objects.create(email="info@trionyx.com")

    def test_task_success(self):
        TestTask().delay(
            task_object=self.user,
            task_user=self.user,
        )

        task = Task.objects.first()

        self.assertEqual(task.result, 'some results')
        self.assertEqual(task.progress_output[0], 'Halfway')

    def test_task_failed(self):
        TestTask().delay(
            task_user=self.user,
        )

        task = Task.objects.first()

        self.assertEqual(task.status, Task.FAILED)
        self.assertEqual(task.result, 'no object')
        self.assertEqual(task.progress_output[0], 'Halfway')



import uuid
import logging
from django.test import TestCase

from trionyx.trionyx.models import Log
from trionyx.log import enable_db_logger

logger = logging.getLogger('trionyx')


class LogTestCase(TestCase):

    def test_log_entry(self):
        with self.settings(TX_DB_LOG_LEVEL=logging.ERROR):
            enable_db_logger()
            log_message = uuid.uuid4().hex
            logger.error(log_message)

            Log.objects.get(message=log_message)

    def test_no_log_entry(self):
        with self.settings(TX_DB_LOG_LEVEL=logging.ERROR):
            enable_db_logger()
            log_message = uuid.uuid4().hex
            logger.debug(log_message)

            self.assertRaises(Log.DoesNotExist, Log.objects.get, message=log_message)



from setuptools import setup
from setuptools import find_packages
from [[name]] import __version__


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name = '[[name]]',
    packages = find_packages(include=['[[name]]', '[[name]].*']),
    include_package_data=True,
    version = __version__,
    description = '[[name]]',
    long_description=readme(),
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Framework :: Django',
    ],
    license='GPL3',
    python_requires='~=3.6',
    install_requires=[],
    extras_require={
        'dev': [
            'Trionyx',
            'colorlog',
            'django-extensions',
            'django-debug-toolbar',
            'Werkzeug',
            'coverage',
            'flake8',
            'pydocstyle',
            'ipython',
        ]
    },
    entry_points={
        'trionyx.app': [
            '[[name]]=[[name]]',
        ],
    }
)



from trionyx.trionyx.models import User
from trionyx.api import serializers


@serializers.register
class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ['id', 'email', 'verbose_name']


from django.test import TestCase
from django.utils import translation

from trionyx import utils


class UtilsTestCase(TestCase):

    def test_random_string(self):
        self.assertNotEqual(utils.random_string(8), utils.random_string(8))
        self.assertEqual(len(utils.random_string(16)), 16)

    def test_get_current_language(self):
        translation.activate('nl-nl')
        self.assertEqual('nl-nl', utils.get_current_language())

    def test_get_local(self):
        translation.activate('nl-nl')
        self.assertEqual('nl_NL', utils.get_current_locale())

    def test_get_datetime_input_format(self):
        translation.activate('nl-nl')
        self.assertEqual('%d-%m-%Y %H:%M:%S', utils.get_datetime_input_format())
        translation.activate('en-us')
        self.assertNotEqual('%d-%m-%Y %H:%M:%S', utils.get_datetime_input_format())

    def test_datetime_format_to_momentjs(self):
        self.assertEqual('DD-MM-YYYY HH:mm:ss', utils.datetime_format_to_momentjs('%d-%m-%Y %H:%M:%S'))

    def test_datetime_format_to_django_template(self):
        self.assertEqual('d-m-Y H:i:s', utils.datetime_format_to_django_template('%d-%m-%Y %H:%M:%S'))



"""Live project settings"""
from .base import *  # noqa: F401,F403
import os


LOG_DIR = get_env_var('LOG_DIR', BASE_DIR)
os.makedirs(LOG_DIR, exist_ok=True)


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': "[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s",
            'datefmt': "%d/%b/%Y %H:%M:%S"
        },
    },
    'handlers': {
        'error-file': {
            'level': 'WARNING',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': os.path.join(LOG_DIR, 'error.log'),
            'maxBytes': 1024*1024*5, # 5 MB
            'backupCount': 10,
            'formatter': 'verbose'
        },
    },
    'loggers': {
        'apps': {
            'level': 'WARNING',
            'handlers': ['error-file'],
        },
        'trionyx': {
            'level': 'WARNING',
            'handlers': ['error-file'],
        },
        'django': {
            'level': 'ERROR',
            'handlers': ['error-file'],
        },
    }
}



from django.test import TestCase
from django.template import Context, RequestContext, Template
from django.utils import timezone
from django.test.client import RequestFactory

from trionyx.trionyx.models import User
from trionyx.menu import MenuItem
from trionyx.layout import HtmlTagWrapper


class TemplateTagTest(TestCase):
    def test_active_menu_item(self):
        request = RequestFactory().get('/dashboard/')
        request.user = User(is_superuser=True)

        context = RequestContext(
            request,
            {
                'item': MenuItem(
                    path='dashboard',
                    name='Dashboard',
                    url='/dashboard/'
                )
            }
        )
        template = Template(
            '{% load trionyx %}'
            '{% active_menu_item request item %}'
        )
        self.assertInHTML('active', template.render(context))

    def test_not_active_menu_item(self):
        request = RequestFactory().get('/not/')
        request.user = User(is_superuser=True)

        context = RequestContext(
            request,
            {
                'item': MenuItem(
                    path='dashboard',
                    name='Dashboard',
                    url='/dashboard/'
                )
            }
        )
        template = Template(
            '{% load trionyx %}'
            '{% active_menu_item request item %}'
        )
        self.assertInHTML('', template.render(context))

    def test_render_component(self):
        context = Context({'component': HtmlTagWrapper()})
        template = Template(
            '{% load trionyx %}'
            '{% render_component component %}'
        )
        self.assertInHTML('<div></div>', template.render(context))

    def test_jsonify(self):
        context = Context({'data': [1, 2, {'key': 'value'}]})
        template = Template(
            '{% load trionyx %}'
            '{{ data|jsonify }}'
        )
        self.assertInHTML('[1, 2, {"key": "value"}]', template.render(context))

    def test_jsonify_queryset(self):
        context = Context({'data': User.objects.all()})
        template = Template(
            '{% load trionyx %}'
            '{{ data|jsonify }}'
        )
        self.assertInHTML('[]', template.render(context))

    def test_model_url(self):
        context = Context({'user': User()})
        template = Template(
            '{% load trionyx %}'
            "{% model_url user 'list' %}"
        )
        self.assertInHTML('/model/trionyx/user/', template.render(context))

    def test_is_date_true(self):
        context = Context({'date': timezone.now()})
        template = Template(
            '{% load trionyx %}'
            "{{ date|is_date }}"
        )
        self.assertInHTML('True', template.render(context))

    def test_is_date_false(self):
        context = Context({'date': 'No date'})
        template = Template(
            '{% load trionyx %}'
            "{{ date|is_date }}"
        )
        self.assertInHTML('False', template.render(context))

    def test_price(self):
        context = Context({'total': 110.25})
        template = Template(
            '{% load trionyx %}'
            "{{ total|price }}"
        )
        self.assertInHTML('US$ 110,25', template.render(context))



import io
from django.test import TestCase

from trionyx.trionyx.models import User
from trionyx.config import models_config


class ModelsTest(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(email='info@trionyx.com', password='top_secret')
        self.test_user = User.objects.create_user(email='test@test.com', password='top_secret')
        self.client.login(email='info@trionyx.com', password='top_secret')

    def get_user_url(self, id=None, action=None):
        return '/'.join(
            filter(None, [
                '/model/trionyx/user',
                str(id) if id else False,
                action
            ])) + '/'

    # List view
    def test_listview(self):
        response = self.client.get('/model/testblog/post/')
        self.assertEqual(response.status_code, 200)

    def test_404_listview(self):
        response = self.client.get('/model/testblog/not-a-model/')
        self.assertEqual(response.status_code, 404)

    def test_403_listview(self):
        self.client.login(email='test@test.com', password='top_secret')
        response = self.client.get('/model/trionyx/user/')
        self.assertEqual(response.status_code, 403)

    def test_ajax_listview(self):
        response = self.client.get('/model/trionyx/user/ajax/')
        data = response.json()

        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 2)

    def test_ajax_listview_filters(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'filters': '[{"field": "email", "operator": "==", "value": "info@trionyx.com"}]'
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 1)

    def test_ajax_listview_invalid_json_filters(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'filters': '{]'
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 2)

    def test_ajax_listview_select_related(self):
        response = self.client.post('/model/testblog/post/ajax/', {
            'selected_fields': 'id,category,category__name'
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['data']['items']), 0)

    def test_ajax_listview_search(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'search': 'info@trionyx.com'
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 1)

    def test_ajax_listview_page(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'page_size': 1,
            'page': 2,
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 1)

    def test_ajax_listview_page_to_small(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'page': -1,
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 2)

    def test_ajax_listview_page_to_big(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'page': 9999,
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 2)

    def test_ajax_listview_custom_fields(self):
        response = self.client.post('/model/trionyx/user/ajax/', {
            'selected_fields': 'id,email',
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 2)
        self.assertEqual(len(data['data']['items'][0]['row_data']), 2)

    def test_ajax_listview_custom_fields_twice(self):
        self.client.post('/model/trionyx/user/ajax/', {
            'selected_fields': 'id,email',
        })
        response = self.client.post('/model/trionyx/user/ajax/', {
            'selected_fields': 'id,email',
        })
        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']['items']), 2)
        self.assertEqual(len(data['data']['items'][0]['row_data']), 2)

    def test_403_ajax_listview(self):
        self.client.login(email='test@test.com', password='top_secret')
        response = self.client.get('/model/trionyx/user/ajax/')
        self.assertEqual(response.status_code, 403)

    def test_listexportview(self):
        response = self.client.post('/model/trionyx/user/download/', {
            'selected_fields': 'id,email',
        })
        content = io.StringIO((b"".join(response.streaming_content)).decode('utf-8'))
        lines = content.readlines()

        self.assertEqual(response.status_code, 200)
        self.assertEqual(lines[0].strip(), 'id,email')
        self.assertEqual(lines[1].strip(), '{},test@test.com'.format(self.test_user.id))
        self.assertEqual(lines[2].strip(), '{},info@trionyx.com'.format(self.user.id))

    def test_listchoices(self):
        response = self.client.get('/model/trionyx/user/choices/', {
            'field': 'created_by',
        })

        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']), 2)

    def test_listchoices_nested(self):
        response = self.client.get('/model/testblog/post/choices/', {
            'field': 'category__created_by',
        })

        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']), 2)

    def test_listchoices_invalid(self):
        response = self.client.get('/model/testblog/post/choices/', {
            'field': 'category__invalid',
        })

        data = response.json()
        self.assertEqual(data['status'], 'success')
        self.assertEqual(len(data['data']), 0)

    # Test detail view
    def test_detailview(self):
        response = self.client.get(self.get_user_url(self.test_user.id))
        self.assertEqual(response.status_code, 200)

    def test_404_detailview(self):
        response = self.client.get('/model/trionyx/user/9999/')
        self.assertEqual(response.status_code, 404)

    def test_tab_detailview(self):
        response = self.client.get(self.get_user_url(self.test_user.id, action='tab'), {
            'tab': 'general'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['status'], 'success')

    def test_no_tab_detailview(self):
        response = self.client.get(self.get_user_url(self.test_user.id, action='tab'), {
            'tab': 'no-tab'
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data['status'], 'error')

    def test_custom_header_button_detailview(self):
        config = models_config.get_config(User)
        config.header_buttons = [
            {
                'label': 'TestCustom header button',
                'url': 'trionyx:model-list',
            }
        ]

        response = self.client.get(self.get_user_url())

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'TestCustom header button')

    def test_layout_view(self):
        response = self.client.get(f'/model/trionyx/user/{self.user.id}/layout/trionyx.user-general/')
        self.assertEqual(response.status_code, 200)

    def test_layout_view_404(self):
        response = self.client.get(f'/model/trionyx/user/{self.user.id}/layout/notexists/')
        self.assertEqual(response.status_code, 404)

    def test_layout_update_complete(self):
        response = self.client.get(f'/model/trionyx/user/{self.user.id}/layout-update/trionyx.user-general/')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()['status'], 'success')

    def test_layout_update_selection(self):
        response = self.client.get(f'/model/trionyx/user/{self.user.id}/layout-update/trionyx.user-general/', data={
            'component': 'all-active-permissions-panel',
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()['status'], 'success')

    # create/update/delete
    def test_create_view(self):
        response = self.client.get('/model/trionyx/user/create/')
        self.assertEqual(response.status_code, 200)

    def test_post_create_view(self):
        response = self.client.post('/model/trionyx/user/create/', {
            'new_password1': 'FwWBuNrZQR',
            'new_password2': 'FwWBuNrZQR',
            'email': 'new@trionyx.com',
        })

        self.assertEqual(response.status_code, 302)
        self.assertEqual(User.objects.count(), 3)

    def test_404_create_view(self):
        response = self.client.get('/model/trionyx/no-model/create/')
        self.assertEqual(response.status_code, 404)

    def test_update_view(self):
        response = self.client.get(self.get_user_url(self.test_user.id, action='edit'))
        self.assertEqual(response.status_code, 200)

    def test_post_update_view(self):
        response = self.client.post(self.get_user_url(self.test_user.id, action='edit'), {
            'email': self.test_user.email,
            'first_name': 'test',
            'is_active': 'True'
        })
        user = User.objects.get(id=self.test_user.id)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(user.first_name, 'test')

    def test_404_update_view(self):
        response = self.client.get('/model/trionyx/user/9999/edit/')
        self.assertEqual(response.status_code, 404)

    def test_delete_view(self):
        response = self.client.get(self.get_user_url(self.test_user.id, action='delete'))
        self.assertEqual(response.status_code, 200)

    def test_post_delete_view(self):
        response = self.client.post(self.get_user_url(self.test_user.id, action='delete'))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(User.objects.count(), 1)

    def test_404_delete_view(self):
        response = self.client.get(self.get_user_url(9999, action='delete'))
        self.assertEqual(response.status_code, 404)



"""
trionyx.views
~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import logging
from collections import defaultdict

from django.utils.translation import ugettext_lazy as _
from trionyx.config import models_config, TX_MODEL_OVERWRITES
from trionyx.layout import Layout, Column12, Panel, DescriptionList, Component

from .models import (  # noqa F401
    ListView, ListJsendView, ListExportView, ListChoicesJsendView, DetailTabView,
    DetailTabJsendView, UpdateView, CreateView, DeleteView, LayoutView, LayoutUpdateView,
)
from .dialogs import (  # noqa F401
    DialogView, UpdateDialog, CreateDialog, LayoutDialog, DeleteDialog
)

from .ajax import JsendView  # noqa F401

logger = logging.getLogger(__name__)


class LayoutRegister:
    """Class where tab layout can be registered"""

    def __init__(self):
        """Init"""
        self.layouts = {}

    def add_layout(self, code, func):
        """Add layout"""
        if code in self.layouts and self.layouts[code]['layout']:
            raise Exception("Layout {} already registered".format(code))

        if code in self.layouts:
            self.layouts[code]['layout'] = func
        else:
            self.layouts[code] = {
                'layout': func,
                'updates': [],
            }

    def register(self, code):
        """Add layout to register"""
        def wrapper(create_layout):
            self.add_layout(code, create_layout)
            return create_layout
        return wrapper

    def add_layout_update(self, code, func):
        """Add update layout function"""
        if code in self.layouts:
            self.layouts[code]['updates'].append(func)
        else:
            self.layouts[code] = {
                'layout': False,
                'updates': [func],
            }

    def update(self, code):
        """Register an update for layout"""
        def wrapper(update_layout):
            self.add_layout_update(code, update_layout)
            return update_layout
        return wrapper

    def get_layout(self, code, object, layout_id=None):
        """Get complete layout for given object"""
        if code not in self.layouts:
            raise Exception('layout does not exist')

        layout_config = self.layouts.get(code)

        layout = layout_config['layout'](object)
        if isinstance(layout, Component):
            layout = Layout(layout)

        if isinstance(layout, list):
            layout = Layout(*layout)

        for update_layout in layout_config['updates']:
            try:
                update_layout(layout, object)
            except Exception as e:
                logger.error('Could not update layout for {}: {}'.format(
                    code,
                    str(e)
                ))

        from trionyx.urls import model_url

        if layout_id:
            layout.id = layout_id

        layout.update_url = model_url(object, 'layout-update', code=code)
        layout.set_object(object)

        return layout


class SidebarRegister:
    """Register sidebars"""

    def __init__(self):
        """Init"""
        self.sidebars = {}

    def register(self, model_alias, code=None):
        """Add sidebar to register

        The function must return a dict that can contain the following data:

        - **title**: Title of the sidebar
        - **content**: Html content to display in sidebar
        - **fixed_content (optional)**: Html content that is fixed displayed under the title
        - **theme (optional)**: Theme of sidebar can be light or dark
        - **hover (optional)**: Sidebar hovers over content instead of pushing it away
        - **actions (optional)**: List[Dict] of actions  that are displayed in dropdown.
            - **label**: Action label
            - **class**: class of action link
            - **url**: Url used for action link
            - **dialog** Action link url is an dialog
            - **dialog_options**: Dialog options
            - **reload**: On dialog success reload sidebar
            - **divider**: Add divider before action
        """
        model_alias = models_config.get_model_name(model_alias) if model_alias else ''

        if (model_alias, code) in self.sidebars:
            raise Exception(f'There is already a sidebar for model: {model_alias} and code: {code}')

        def wrapper(create_sidebar):
            self.sidebars[(model_alias, code)] = create_sidebar
            return create_sidebar

        return wrapper

    def get_sidebar(self, model_alias, code=None):
        """Get sidebar"""
        model_alias = models_config.get_model_name(model_alias) if model_alias else ''

        if (model_alias, code) not in self.sidebars:
            raise Exception(f'Sidebar does not exists for model: {model_alias} and code: {code}')

        return self.sidebars[(model_alias, code)]


class TabRegister:
    """Class where tab layout can be registered"""

    def __init__(self):
        """Init Tabs"""
        self.tabs = defaultdict(list)

    def get_tabs(self, model_alias, object):
        """
        Get all active tabs for given model

        :param model_alias:
        :param object: Object used to filter tabs
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        for item in self.tabs.get(model_alias, list()):
            if item.display_filter(object):
                yield item

    def get_tab(self, model_alias, object, tab_code):
        """
        Get tab for given object and tab code

        :param model_alias:
        :param object: Object used to render tab
        :param tab_code: Tab code to use
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        for item in self.tabs[model_alias]:
            if item.code == tab_code and item.display_filter(object):
                return item
        raise Exception('Given tab does not exits or is filtered')

    def register(self, model_alias, code='general', name=None, order=10, display_filter=None):
        """
        Register new tab

        :param model_alias:
        :param code:
        :param name:
        :param order:
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        if code == 'general' and not name:
            name = _('General')

        def wrapper(create_layout):
            item = TabItem(
                code=code,
                name=name,
                order=order,
                display_filter=display_filter,
                model_alias=model_alias
            )

            if item in self.tabs[model_alias]:
                raise Exception("Tab {} already registered for model {}".format(code, model_alias))

            self.tabs[model_alias].append(item)
            self.tabs[model_alias] = sorted(self.tabs[model_alias], key=lambda item: item.order if item.order else 10)
            layouts.add_layout(f'{model_alias}-{code}', create_layout)

            return create_layout
        return wrapper

    def register_update(self, model_alias, code):
        """
        Register tab update function, function is being called with (layout, object)

        :param model_alias:
        :param code:
        :return:
        """
        model_alias = self.get_model_alias(model_alias)

        def wrapper(update_layout):
            layouts.add_layout_update(f'{model_alias}-{code}', update_layout)
            return update_layout
        return wrapper

    def update(self, model_alias, code='general', name=None, order=10, display_filter=None):
        """
        Update given tab

        :param model_alias:
        :param code:
        :param name:
        :param order:
        :param display_filter:
        :return:
        """
        model_alias = self.get_model_alias(model_alias)
        for item in self.tabs[model_alias]:
            if item.code != code:
                continue
            if name:
                item.name = name
            if order:
                item.order = order
            if display_filter:
                item.display_filter = display_filter
            break
        self.tabs[model_alias] = sorted(self.tabs[model_alias], key=lambda item: item.code if item.code else 10)

    def get_model_alias(self, model_alias):
        """Get model alias if class then convert to alias string"""
        name = models_config.get_model_name(model_alias)
        return TX_MODEL_OVERWRITES.get(name, name)

    def auto_generate_missing_tabs(self):
        """Auto generate tabs for models with no tabs"""
        for config in models_config.get_all_configs(False):
            model_alias = self.get_model_alias(config.model)
            if model_alias not in self.tabs:
                @self.register(model_alias, order=10)
                def general_layout(obj):
                    return Layout(
                        Column12(
                            Panel(
                                'info',
                                DescriptionList(*[f.name for f in models_config.get_config(obj).get_fields()])
                            )
                        )
                    )


class TabItem:
    """Tab item that holds the tab data and renders the layout"""

    def __init__(self, code, name=None, order=None, display_filter=None, model_alias=None):
        """Init TabItem"""
        self._name = None
        self.code = code
        self.name = name
        self.order = order
        self.display_filter = display_filter if display_filter else lambda object: True
        self.model_alias = model_alias

    @property
    def name(self):
        """Give back tab name if is set else generate name by code"""
        if self._name:
            return self._name
        return self.code.replace('_', ' ').capitalize()

    @name.setter
    def name(self, name):
        """Set name"""
        self._name = name

    def get_layout(self, object):
        """Get complete layout for given object"""
        model_alias = self.model_alias if self.model_alias else tabs.get_model_alias(object)
        return layouts.get_layout(f'{model_alias}-{self.code}', object)

    def __str__(self):
        """Tab string representation"""
        if not self.name:
            return self.name.capitalize()
        return self.name

    def __eq__(self, other):
        """Compare tab based on code"""
        return self.code == other.code


tabs = TabRegister()
layouts = LayoutRegister()
sidebars = SidebarRegister()



from trionyx import forms

from app.testblog.models import Tag, Post, Category


@forms.register(code='new', default_edit=True)
class TagForm(forms.ModelForm):
    name = forms.CharField(label='name (default edit form)')

    class Meta:
        model = Tag
        fields=['name']


@forms.register(default_edit=True)
class PostForm(forms.ModelForm):
    category = forms.ModelAjaxChoiceField(Category.objects.all())
    tags = forms.ModelAjaxMultipleChoiceField(Tag.objects.all(), required=False)

    class Meta:
        model = Post
        fields = ['title', 'content', 'publish_date', 'category', 'status', 'price', 'tags']


"""
trionyx.views.ajax
~~~~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
import logging
from typing import ClassVar, Any

from django.http import JsonResponse
from django.http.request import HttpRequest
from django.views.generic import View

logger = logging.getLogger(__name__)


class JsendView(View):
    """Django view for sending JSON response by the JSend specification

    Specification website: http://labs.omniti.com/labs/jsend
    """

    SUCCESS: ClassVar[str] = 'success'
    ERROR: ClassVar[str] = 'error'
    FAIL: ClassVar[str] = 'fail'
    allow_cors: bool = False

    def __init__(self, *args, **kwargs):
        """Init JsendView"""
        super(JsendView, self).__init__(*args, **kwargs)
        self.status: str = self.SUCCESS
        self.data: Any = ""
        self.message: str = ""

    def handle_request(self, request: HttpRequest, *args, **kwargs):
        """Overide this in your class for handling the request"""
        pass

    def _handle_request(self, request: HttpRequest, *args, **kwargs):
        """Handle request"""
        try:
            data = self.handle_request(request, *args, **kwargs)
        except Exception as e:
            logger.exception(e)
            self.status = self.ERROR
            self.message = str(e)

        if self.status == self.ERROR:
            jsend = {
                'status': self.status,
                'message': self.message
            }
        else:
            jsend = {
                'status': self.status,
                'data': data
            }
        response = JsonResponse(jsend)
        if self.allow_cors:
            response['Access-Control-Allow-Origin'] = "*"
        return response

    def get(self, request: HttpRequest, *args, **kwargs):
        """Handle GET request"""
        return self._handle_request(request, *args, **kwargs)

    def post(self, request: HttpRequest, *args, **kwargs):
        """Handle POST request"""
        return self._handle_request(request, *args, **kwargs)



from django.test import TestCase

from trionyx.trionyx.models import User


class ModelsTest(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(email='info@trionyx.com', password='top_secret')
        self.test_user = User.objects.create_user(email='test@test.com', password='top_secret')
        self.client.login(email='info@trionyx.com', password='top_secret')

    def get_user_url(self, id, action=None):
        return '/'.join(
            filter(None, [
                '/dialog/model/trionyx/user',
                str(id),
                action
            ])) + '/'

    # Update Dialog
    def test_update_dialog(self):
        response = self.client.get(self.get_user_url(self.test_user.id, action='edit'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.test_user.email)

    def test_post_update_dialog(self):
        response = self.client.post(self.get_user_url(self.test_user.id, action='edit'), {
            'email': self.test_user.email,
            'first_name': 'test',
            'is_active': 'True'
        })
        user = User.objects.get(id=self.test_user.id)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(user.first_name, 'test')
        self.assertContains(response, 'alert-success')

    def test_404_update_dialog(self):
        response = self.client.get('/dialog/model/trionyx/user/9999/edit/')
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Could not load')

    # Create dialog
    def test_create_dialog(self):
        response = self.client.get('/dialog/model/trionyx/user/create/')
        self.assertEqual(response.status_code, 200)

    def test_post_create_dialog(self):
        response = self.client.post('/dialog/model/trionyx/user/create/', {
            'new_password1': 'FwWBuNrZQR',
            'new_password2': 'FwWBuNrZQR',
            'email': 'new@trionyx.com',
        })

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'alert-success')
        self.assertEqual(User.objects.count(), 3)

    def test_404_create_dialog(self):
        response = self.client.get('/dialog/model/trionyx/no-model/create/')
        self.assertEqual(response.status_code, 200)
        # TODO Add better error message

    # Delete dialog
    def test_delete_dialog(self):
        response = self.client.get(self.get_user_url(self.test_user.id, action='delete'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.test_user.email)

    def test_post_delete_dialog(self):
        response = self.client.post(self.get_user_url(self.test_user.id, action='delete'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, '"success": true')
        self.assertEqual(User.objects.count(), 1)

    def test_404_delete_dialog(self):
        response = self.client.get(self.get_user_url(9999, action='delete'))
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Could not load')



from trionyx.views import tabs, layouts, sidebars
from trionyx.layout import *
from trionyx.urls import model_url
from trionyx import models
from django.db.models.functions import TruncDay
from trionyx.renderer import price_value_renderer

from .models import Post, Tag

@layouts.register('view')
@tabs.register(Post)
def post_overview(obj):
    return [
        Column6(
            Panel(
                "Post info",
                TableDescription(
                    'status',
                    'title',
                    'publish_date',
                    'category',
                    'tags',
                    'price',
                    {
                        'label': 'Update',
                        'value': Button(
                            'Update',
                            url=model_url(obj, 'dialog-edit'),
                            dialog=True,
                            # dialog_options={
                            #     'callback': "function(){trionyx_reload_tab('general')}"
                            # },
                            dialog_reload_layout='table-description',
                            color=Colors.TEAL,
                        )
                    },
                    {
                        'label': 'Sidebar',
                        'value': Button(
                            'Sidebar',
                            model_url='sidebar',
                            model_code='item',
                            sidebar=True,
                        )
                    },
                )
            ),
            Panel(
                'Charts',
                LineChart(
                    Post.objects.order_by('-sale_date'),
                    'sale_date',
                    'price',
                ),
                LineChart(
                    Post.objects.annotate(day=TruncDay('publish_date')).values(
                        'day').annotate(sum_price=models.Sum('price')).values('day', 'sum_price').order_by('-day'),
                    'day',
                    {
                        'label': 'Total price per day',
                        'field': 'sum_price',
                        'renderer': price_value_renderer,
                    },
                    time_unit='day',
                ),
                LineChart(
                    'tags',
                    'name',
                    'id',
                ),
                BarChart(
                    Post.objects.annotate(day=TruncDay('publish_date')).values('day'
                    ).annotate(
                        sum_price=models.Sum('price'),
                        sum_id=models.Sum('id')).values('day', 'sum_price', 'sum_id').order_by('-day'),
                    'day',
                    {
                        'label': 'Total price per day',
                        'field': 'sum_price',
                        'renderer': price_value_renderer,
                    },
                    'sum_id',
                    time_unit='day',
                ),
                DoughnutChart(
                    [['Python3', 60], ['Javascript', 30], ['Sql', 10]],
                    'name',
                    'value',
                ),
                PieChart(
                    [['Python3', 60], ['Javascript', 30], ['Sql', 10]],
                    'name',
                    'value',
                ),
            )
        ),
        Column6(
            Panel(
                'Components',
                TableDescription(
                    {
                        'label': 'img',
                        'value': Img(
                            src='https://www.python.org/static/img/python-logo.png',
                            width='200px',
                        ),
                    },
                    {
                        'label': 'Input',
                        'value': Input(name='test')
                    },
                    {
                        'label': 'Badges',
                        'value': Component(
                            Badge('status', color=Colors.THEME),
                            Badge('status', color=Colors.LIGHT_BLUE),
                            Badge('status', color=Colors.AQUA),
                            Badge('status', color=Colors.GREEN),
                            Badge('status', color=Colors.YELLOW),
                            Badge('status', color=Colors.RED),
                            Badge('status', color=Colors.GRAY),
                            Badge('status', color=Colors.NAVY),
                            Badge('status', color=Colors.TEAL),
                            Badge('status', color=Colors.PURPLE),
                            Badge('status', color=Colors.ORANGE),
                            Badge('status', color=Colors.MAROON),
                            Badge('status', color=Colors.BLACK),
                        )
                    },
                    {
                        'label': 'Alerts',
                        'value': Component(
                            Alert('Warning with no_margin', alert=Alert.INFO,no_margin=True),
                            Alert('success', alert=Alert.SUCCESS),
                            Alert('warning', alert=Alert.WARNING),
                            Alert('danger', alert=Alert.DANGER),
                        )
                    },
                    {
                        'label': 'Thumbnail',
                        'value': Thumbnail(
                            'https://www.python.org/static/img/python-logo.png',
                            url=model_url(obj, 'dialog-edit'),
                            dialog=True,
                            dialog_options={
                                'callback': "function(){trionyx_reload_tab('general')}"
                            },
                        )
                    },
                    {
                        'label': 'ProgressBar',
                        'value': Component(
                            ProgressBar(value=543, max_value=2400),
                            ProgressBar(value=20, size='xs'),
                            ProgressBar(value=78, active=True, color=Colors.GREEN),
                        ),
                    },
                    {
                        'label': 'Lists',
                        'value': UnorderedList(
                            'title',
                            'content',
                            {
                                'value': 'Fixed value'
                            },
                            {
                                'label': 'Unordered sublist',
                                'value': UnorderedList(
                                    'status',
                                    {
                                        'label': 'Another unordered sublist',
                                        'value': UnorderedList(
                                            'price'
                                        ),
                                    }
                                )
                            },
                            {
                                'label': 'Ordered sublist',
                                'value': OrderedList(
                                    'status',
                                    {
                                        'label': 'Another ordered sublist',
                                        'value': OrderedList(
                                            'price'
                                        ),
                                    }
                                )
                            },
                            {
                                'label': 'tags',
                                'value': UnorderedList(
                                    'name',
                                    objects='tags'
                                ),
                            }
                        )
                    },
                    id="table-description"
                ),
            )
        ),
        Column12(
            Panel(
                'Tags',
                Table(
                    [
                        ['list object'],
                        {
                            'name': 'dict object',
                        },
                        Tag(name='Tag object'),
                    ],
                    'name=width:120px',
                    'Value=value:fixed value',
                    {
                        'label': 'Value',
                        'value': 'fixed value',
                        'renderer': lambda value, data_object, **options: '{}'.format(data_object),
                    },
                    footer=[
                        [
                            ['Subtotaal', '12,50'],
                            ['Totaal', '12,50'],
                        ],
                        {
                            'colspan': 2,
                            'class': 'text-right'
                        },
                        'Price'
                    ],
                    hover=True,
                ),
                collapse=True,
            )
        )
    ]

@tabs.register_update('trionyx.user', 'general')
def tasks_update(layout, obj):
    layout.add_component(Panel('Inserted'), path='container.row.column10[1].panel')


@sidebars.register(Post, 'item')
def item_sidebar(request, obj):
    layout = post_overview(obj)


    return {
        'title': str(obj),
        'fixed_content': 'Some fixed content 2',
        'content': 'Real content <br>' * 100,
        'theme': 'light',
        'hover': False,
        'actions': [
            {
                'label': 'Save',
                'class': 'text-success text-bold',
                'url': model_url(obj, 'dialog-edit'),
                'dialog': True,
                'dialog_options': {},
                'reload': True,
            },
            {
                'label': 'Warning',
                'class': 'text-warning',
                'url': model_url(obj, 'dialog-edit'),
                'dialog': True,
                'dialog_options': {},
                'reload': True,
            },
            {
                'label': 'Info',
                'class': 'text-info',
                'url': model_url(obj, 'dialog-edit'),
                'dialog': True,
                'dialog_options': {},
                'reload': True,
            },
            {
                'label': 'Delete',
                'class': 'text-danger',
                'url': model_url(obj, 'dialog-edit'),
                'dialog': True,
                'dialog_options': {},
                'reload': True,
                'divider': True,
            }
        ]
    }


from django.test import TestCase

from trionyx.menu import Menu, MenuItem


class UtilsTestCase(TestCase):

    def test_merge_menu_item(self):
        item1 = MenuItem(
            path='item1',
            name='item1',
            icon='item1',
            url='item1',
            order=10,
            permission='item1'
        )
        item2 = MenuItem(
            path='item1',
            name='item2',
            icon='item2',
            order=20,
        )

        item1.merge(item2)
        self.assertEquals(item1.path, 'item1')
        self.assertEquals(item1.name, 'item2')
        self.assertEquals(item1.icon, 'item2')
        self.assertEquals(item1.order, 20)

    def test_complete_merge_menu_item(self):
        item1 = MenuItem(
            path='item1',
            name='item1',
            icon='item1',
            url='item1',
            order=10,
            permission='item1'
        )
        item2 = MenuItem(
            path='item1',
            name='item2',
            icon='item2',
            url='item2',
            order=20,
            permission='item2'
        )

        item1.merge(item2)
        self.assertEquals(item1.path, 'item1')
        self.assertEquals(item1.name, 'item2')
        self.assertEquals(item1.icon, 'item2')
        self.assertEquals(item1.url, 'item2')
        self.assertEquals(item1.order, 20)
        self.assertEquals(item1.permission, 'item2')

    def test_add_menu_item(self):
        menu = Menu()
        menu.add_item('/test', 'test')
        items = menu.get_menu_items()

        self.assertEquals(len(items), 1)
        self.assertEquals(items[0].name, 'test')

    def test_add_sub_menu_path(self):
        menu = Menu()
        menu.add_item('/test/sub', 'sub')
        items = menu.get_menu_items()

        # Test first item
        self.assertEquals(len(items), 1)
        self.assertEquals(items[0].name, 'Test')
        self.assertEquals(items[0].path, '/test')
        self.assertEquals(len(items[0].childs), 1)

        # Test sub item
        self.assertEquals(items[0].childs[0].name, 'sub')
        self.assertEquals(items[0].childs[0].path, '/test/sub')

    def test_add_same_menu_sub_path(self):
        menu = Menu()
        menu.add_item('/test/sub', 'sub')
        menu.add_item('/test/sub2', 'sub2')
        items = menu.get_menu_items()

        self.assertEquals(len(items), 1)

        # Test first item
        self.assertEquals(items[0].name, 'Test')
        self.assertEquals(items[0].path, '/test')
        self.assertEquals(len(items[0].childs), 2)

        # Test first sub item
        self.assertEquals(items[0].childs[0].name, 'sub')
        self.assertEquals(items[0].childs[0].path, '/test/sub')

        # Test first sub item
        self.assertEquals(items[0].childs[1].name, 'sub2')
        self.assertEquals(items[0].childs[1].path, '/test/sub2')

    def test_add_menu_merge(self):
        menu = Menu()
        menu.add_item('/test', 'test')
        menu.add_item('/test', 'New name', order=10)
        items = menu.get_menu_items()

        self.assertEquals(len(items), 1)
        self.assertEquals(items[0].name, 'New name')
        self.assertEquals(items[0].path, '/test')
        self.assertEquals(items[0].order, 10)

    def test_add_menu_item_order(self):
        menu = Menu()
        menu.add_item('/first', 'first', order=10)
        menu.add_item('/third', 'third')
        menu.add_item('/second', 'second', order=20)
        items = menu.get_menu_items()

        self.assertEquals(len(items), 3)
        self.assertEquals(items[0].name, 'first')
        self.assertEquals(items[1].name, 'second')
        self.assertEquals(items[2].name, 'third')

    def test_menu_item_active(self):
        menu = Menu()
        menu.add_item('/item', 'item', url='/item/')
        menu.add_item('/item/subitem', 'item', url='/item/subitem/')
        item = menu.get_menu_items()[0]

        self.assertFalse(item.is_active('/'))
        self.assertFalse(item.is_active('/items/'))
        self.assertTrue(item.is_active('/item/'))
        self.assertTrue(item.is_active('/item/subitem/'))

    def test_root_menu_item_active(self):
        menu = Menu()
        menu.add_item('root', 'root', url='/')
        item = menu.get_menu_items()[0]

        self.assertTrue(item.is_active('/'))
        self.assertFalse(item.is_active('/path/'))

    def test_regex_menu_item_active(self):
        menu = Menu()
        menu.add_item('regex', 'regex', active_regex=r'/test.+/')
        item = menu.get_menu_items()[0]

        self.assertFalse(item.is_active('/test/'))
        self.assertTrue(item.is_active('/test123/'))



"""
trionyx.quickstart
~~~~~~~~~~~~~~~~~~

Quickstart Trionyx project and apps

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import os
import shutil
import subprocess

import yaml
import trionyx
from trionyx import utils


class Quickstart:
    """Class for creating Trionyx project and apps"""

    def __init__(self):
        """Set project template paths"""
        quickstart_path = os.path.realpath(__file__)

        self.project_path = os.path.join(os.path.dirname(quickstart_path), 'project')
        """Path to project template files"""

        self.app_path = os.path.join(os.path.dirname(quickstart_path), 'app')
        """Path to app template files"""

        self.reusable_app_path = os.path.join(os.path.dirname(quickstart_path), 'reusable_app')
        """Path to reusable app template files"""

        self.ansible_path = os.path.join(os.path.dirname(quickstart_path), 'ansible')
        """Path to Ansible template files"""

    def create_project(self, project_path):
        """
        Create Trionyx project in given path

        :param str path: path to create project in.
        :raises FileExistsError:
        """
        shutil.copytree(self.project_path, project_path)

        self.update_file(project_path, 'requirements.txt', {
            'trionyx_version': trionyx.__version__
        })

        self.update_file(project_path, 'environment.json', {
            'secret_key': utils.random_string(32)
        })

        app_name = os.path.basename(project_path).capitalize()

        self.update_file(project_path, 'README.rst', {
            'title': "{name}\n{heading}".format(
                name=app_name,
                heading='=' * len(app_name),
            )
        })

        self.update_file(project_path, 'config/settings/base.py', {
            'app_name': app_name,
            'logo_name_start': app_name[0],
            'logo_name_end': app_name[1:],
            'logo_name_small_start': app_name[0].upper(),
            'logo_name_small_end': app_name[1].upper(),
        })

    def create_app(self, apps_path, name):
        """
        Create Trionyx app in given path

        :param str path: path to create app in.
        :param str name: name of app
        :raises FileExistsError:
        """
        app_path = os.path.join(apps_path, name.lower())

        shutil.copytree(self.app_path, app_path)

        self.update_file(app_path, '__init__.py', {
            'name': name.lower()
        })

        self.update_file(app_path, 'apps.py', {
            'name': name.lower(),
            'verbose_name': name.capitalize()
        })

    def create_reusable_app(self, path, name):
        """
        Create Trionyx reusable app

        :param path:
        :param name:
        :return:
        """
        shutil.copytree(self.reusable_app_path, path)

        variables = {
            'name': name.lower(),
            'verbose_name': name.capitalize(),
        }

        self.update_file(path, '[app_name]/apps.py', variables)
        self.update_file(path, '[app_name]/__init__.py', variables)
        self.update_file(path, 'app/settings.py', variables)
        self.update_file(path, 'MANIFEST.in', variables)
        self.update_file(path, 'README.rst', variables)
        self.update_file(path, 'setup.py', variables)

        # Rename package
        shutil.move(os.path.join(path, '[app_name]'), os.path.join(path, name.lower()))

    def create_ansible(self, project_path, domain, repo):
        """Create Ansible live deploy script"""
        shutil.copytree(self.ansible_path, project_path)

        self.update_file(project_path, 'production', {
            'domain': domain,
        })

        os.mkdir(os.path.join(project_path, 'ssh_keys'))
        subprocess.check_output([
            'ssh-keygen', '-t', 'rsa', '-b', '4096', '-N', '', '-f', os.path.join(project_path, 'ssh_keys', 'deploy_rsa')
        ])
        subprocess.check_output([
            'ssh-keygen', '-t', 'rsa', '-b', '4096', '-N', '', '-f', os.path.join(project_path, 'ssh_keys', 'connect_rsa')
        ])

        with open(os.path.join(project_path, 'ssh_keys', 'deploy_rsa')) as _file:
            deploy_key = _file.read()

        config = {
            'ansible_ssh_user': 'ansible',
            'ansible_ssh_port': 6969,
            'ansible_ssh_private_key_file': 'ssh_keys/connect_rsa',
            'app_domain': domain,
            'app_repo': repo,
            'app_config_template': "{{ playbook_dir }}/templates/environment.json.j2",
            'deploy_key': deploy_key,
            'secret_key': utils.random_string(32),
            'db_password': utils.random_string(16),
            'broker_password': utils.random_string(16),
            'smtp_password': '',
        }

        self.update_file(project_path, 'group_vars/all.yml', {
            'config': yaml.dump(config, default_flow_style=False, allow_unicode=True),
        })

    def update_file(self, project_path, file_path, variables):
        """
        Update given file with given variables, variables in file must be inclosed with [[]].

        For example you want to replace a variable secret_key, in the file its [[secret_key]].

        :param str project_path:
        :param str file_path:
        :param dict variables:
        """
        update_file = os.path.join(project_path, file_path)
        with open(update_file, 'rb') as _file:
            file_content = _file.read().decode('utf-8')

        for key, value in variables.items():
            file_content = file_content.replace('[[{}]]'.format(key), value)

        with open(update_file, 'w+', encoding='utf-8') as _file:
            _file.writelines(file_content)



"""
trionyx.trionyx.management.commands.create_favicon
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import os

from PIL import Image, ImageDraw, ImageFont
from django.core.management.base import BaseCommand
from django.conf import settings


class Command(BaseCommand):
    """Command for creating favicon"""

    help = 'Create favicon'

    def handle(self, *args, **options):
        """Create new favicon"""
        app_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        skins = {
            'blue': (0, 115, 183),
            'yellow': (243, 156, 18),
            'green': (0, 166, 90),
            'purple': (96, 92, 168),
            'red': (221, 75, 57),
            'black': (17, 17, 17),
        }
        theme = settings.TX_THEME_COLOR.replace('-light', '')
        color = skins.get(theme, (17, 17, 17))

        text = '{}{}'.format(
            settings.TX_LOGO_NAME_SMALL_START.upper(),
            settings.TX_LOGO_NAME_SMALL_END.lower()
        )

        img = Image.new('RGB', (32, 32), color=color)
        draw = ImageDraw.Draw(img)
        font = ImageFont.truetype(os.path.join(app_dir, 'static', 'fonts', 'pt-mono.bold.ttf'), 24)

        text_width, text_height = draw.textsize(text, font)
        position = ((32 - text_width) / 2, (32 - 5 - text_height) / 2)
        draw.text(position, text, (255, 255, 255), font=font)
        img.save(os.path.join(app_dir, 'static', 'img', 'favicon.png'))



"""
trionyx.trionyx.view.accounts
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import os
import re
import json
import logging
from collections import OrderedDict

from docutils.core import publish_parts
from django.core.paginator import Paginator
from django.http import HttpResponse, HttpResponseRedirect, JsonResponse
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.contrib.auth import logout as django_logout
from django.contrib.auth.models import Permission
from django.shortcuts import redirect
from django.urls import reverse
from watson import search as watson
from django.contrib.contenttypes.models import ContentType
from django.views.generic import TemplateView
from django.conf import settings
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.utils.translation import ugettext_lazy as _
from django.contrib import messages
from django.forms.widgets import SelectMultiple

from trionyx.models import get_class
from trionyx.views import UpdateView, DetailTabView, DialogView, JsendView
from trionyx.views.mixins import ModelClassMixin, ModelPermissionMixin
from trionyx.config import models_config
from trionyx.widgets import widgets
from trionyx import utils
from trionyx.forms.helper import FormHelper
from trionyx.forms import form_register, modelform_factory, ModelAjaxChoiceField
from trionyx.models import filter_queryset_with_user_filters
from trionyx.trionyx.tasks import MassUpdateTask

User = get_class('trionyx.User')
Task = get_class('trionyx.Task')

logger = logging.getLogger(__name__)


def basic_auth(request, user_type):
    """View that can be used for basic-auth like Nginx ngx_http_auth_request_module"""
    if user_type == 'superuser' and request.user.is_authenticated and request.user.is_superuser:
        return HttpResponse('OK')
    if user_type == 'user' and request.user.is_authenticated:
        return HttpResponse('OK')
    return HttpResponse('NOK', status=403)


def media_nginx_accel(request, path):
    """
    Location /protected/ {
        internal;
        alias <complete path to project root dir>;
    }

    """
    response = HttpResponse(status=200)
    response['Content-Type'] = ''
    response['X-Accel-Redirect'] = '/protected' + request.path
    return response


class LoginView(DjangoLoginView):
    """Trionyx login view"""

    redirect_authenticated_user = True

    template_name = 'trionyx/core/login.html'


def logout(request):
    """Trionyx logout view"""
    django_logout(request)
    return redirect('/')


class UpdateUserAccountView(UpdateView):
    """Update user view"""

    model = User
    title = 'Update account'
    cancel_url = 'trionyx:view-account'

    @property
    def form_class(self):
        """Form class used by view"""
        from trionyx.trionyx.forms import ProfileUpdateForm
        return ProfileUpdateForm

    def get(self, request, *args, **kwargs):
        """Add user id to kwargs"""
        kwargs['pk'] = request.user.id
        self.kwargs['pk'] = request.user.id
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """Add user id to kwargs"""
        kwargs['pk'] = request.user.id
        self.kwargs['pk'] = request.user.id
        return super().post(request, *args, **kwargs)

    def get_success_url(self):
        """Set success url to account page"""
        return reverse('trionyx:view-account')


class ViewUserAccountView(DetailTabView):
    """View user"""

    model_alias = 'trionyx.profile'
    model = User

    title = 'Profile'

    def get(self, request, *args, **kwargs):
        """Add user id to kwargs"""
        kwargs['pk'] = request.user.id
        self.kwargs['pk'] = request.user.id
        return super().get(request, *args, **kwargs)

    def get_edit_url(self):
        """Get user edit url"""
        return reverse('trionyx:edit-account')

    def get_delete_url(self):
        """Hide delete url"""
        return None

    def get_back_url(self):
        """Hide back url"""
        return None


def create_permission_jstree(selected=None, disabled=False):
    """Create permission jstree"""
    selected = selected if selected else []
    jstree = []

    added_apps = ['auth']
    added_models = []
    for permission in Permission.objects.select_related('content_type').all():
        if not permission.content_type.model_class():
            continue

        model_config = models_config.get_config(permission.content_type.model_class())

        if model_config.hide_permissions:
            continue

        if model_config.disable_add and permission.codename == 'add_{}'.format(model_config.model_name):
            continue

        if model_config.disable_change and permission.codename == 'change_{}'.format(model_config.model_name):
            continue

        if model_config.disable_delete and permission.codename == 'delete_{}'.format(model_config.model_name):
            continue

        parent = ['jstree']

        if model_config.app_label not in added_apps:
            jstree.append({
                'id': '.'.join([*parent, model_config.app_label]),
                'parent': '#',
                'text': model_config.get_app_verbose_name(),
                'state': {
                    'disabled': disabled,
                }
            })
            added_apps.append(model_config.app_label)

        parent.append(model_config.app_label if model_config.app_label != 'auth' else 'trionyx')

        if model_config.model_name not in added_models:
            jstree.append({
                'id': '.'.join([*parent, model_config.model_name]),
                'parent': '.'.join(parent),
                'text': model_config.get_verbose_name_plural(),
                'state': {
                    'disabled': disabled,
                }
            })
            added_models.append(model_config.model_name)

        parent.append(model_config.model_name)
        name = {
            'view_{}'.format(model_config.model_name): _('View'),
            'add_{}'.format(model_config.model_name): _('Add'),
            'change_{}'.format(model_config.model_name): _('Change'),
            'delete_{}'.format(model_config.model_name): _('Delete'),
        }.get(permission.codename, permission.name)

        jstree.append({
            'id': '.'.join([*parent, permission.codename]),
            'parent': '.'.join(parent),
            'text': str(name),
            'state': {
                'selected': permission in selected,
                'disabled': disabled,
            },
            'permission_id': permission.id,
        })

    return jstree


# =============================================================================
# Ajax form views
# =============================================================================
def ajaxFormModelChoices(request, id):
    """View for select2 ajax data request"""
    if id not in ModelAjaxChoiceField.registered_fields:
        return JsonResponse({})

    field = ModelAjaxChoiceField.registered_fields[id]
    query = field.queryset.order_by('verbose_name')

    search = request.GET.get('q', None)
    if search:
        config = models_config.get_config(query.model)
        if not config.disable_search_index:
            query = watson.filter(query, search, ranking=False)
        else:
            query = query.filter(verbose_name__contains=search)

    pages = Paginator(query, 20)
    page = pages.page(int(request.GET.get('page', 1)))

    result = [
        {'id': '', 'text': '------'}
    ] if not isinstance(field.widget, SelectMultiple) and not field.required and page.number == 1 else []

    return JsonResponse({
        'results': result + [{
            'id': row[0],
            'text': row[1],
        } for row in page.object_list.values_list('id', 'verbose_name')],
        'pagination': {
            'more': page.has_next(),
        }
    })


# =============================================================================
# Global search
# =============================================================================
class GlobalSearchJsendView(JsendView):
    """View for global search uses watson to search all models"""

    def handle_request(self, request):
        """Handle search"""
        models = []
        content_types = {}
        for config in models_config.get_all_configs(False):
            if config.disable_search_index or not config.global_search:
                continue

            # Check if user has view permission
            if not request.user.has_perm('{app_label}.view_{model_name}'.format(
                app_label=config.app_label,
                model_name=config.model_name,
            ).lower()):
                continue

            models.append(config.model)
            content_type = ContentType.objects.get_for_model(config.model, False)
            content_types[content_type.id] = str(config.model._meta.verbose_name_plural)

        results = OrderedDict()
        for entry in watson.search(request.GET.get('search', ''), models=models)[:100]:
            if entry.content_type_id not in results:
                results[entry.content_type_id] = {
                    'name': content_types[entry.content_type_id],
                    'items': [],
                }

            if len(results[entry.content_type_id]['items']) <= 10:
                results[entry.content_type_id]['items'].append({
                    'url': entry.url,
                    'title': entry.title,
                    'description': entry.description,
                })

        return list(results.values())


class FilterFieldsJsendView(JsendView):
    """View for getting model filter fields"""

    def handle_request(self, request, *args, **kwargs):
        """Get filter fields"""
        try:
            modelClass = ContentType.objects.get_for_id(request.GET.get('id')).model_class()
        except ContentType.DoesNotExist:
            return {}
        config = models_config.get_config(modelClass)

        return {
            'id': request.GET.get('id'),
            'fields': {
                name: {
                    'name': name,
                    'label': str(field['label']),
                    'type': field['type'],
                    'choices': field['choices'],
                    'choices_url': field.get('choices_url', None)
                }
                for name, field in config.get_list_fields().items()
            }
        }


class UserTasksJsend(JsendView):
    """User tasks view"""

    def handle_request(self, request):
        """Get user open tasks"""
        tasks = []

        tasks.extend([
            task for task in Task.objects.filter(user=request.user, status=Task.SCHEDULED).order_by('-scheduled_at')
        ])

        tasks.extend([
            task for task in Task.objects.filter(
                user=request.user,
                status__in=[Task.QUEUE, Task.LOCKED, Task.RUNNING]
            ).order_by('-started_at')
        ])

        if len(tasks) < 10:
            tasks.extend([
                task for task in Task.objects.filter(
                    user=request.user, status__in=[Task.COMPLETED, Task.FAILED]
                ).order_by('-started_at')[: 10 - len(tasks)]
            ])

        return [
            {
                'id': task.id,
                'status': task.status,
                'status_display': task.get_status_display(),
                'description': task.description,
                'progress': task.progress,
                'url': task.get_absolute_url(),
            } for task in tasks
        ]


# =============================================================================
# Changelog
# =============================================================================
class ChangelogDialog(DialogView):
    """Dialog to show app changelog based on CHANGELOG.rst"""

    def display_dialog(self):
        """Display changelog"""
        changelog_path = os.path.join(settings.BASE_DIR, 'CHANGELOG.rst')

        if os.path.isfile(changelog_path):
            with open(changelog_path, 'r', encoding='utf-8') as _file:
                content = publish_parts(_file.read(), writer_name='html')['html_body']
        else:
            content = ''

        if settings.TX_CHANGELOG_HASHTAG_URL:
            link_re = r'<a href="{link}" target="_blank">#\1</a>'.format(link=settings.TX_CHANGELOG_HASHTAG_URL.format(tag='\\1'))
            content = re.sub(r"#([\d\w\-]+)", link_re, content)

        return {
            'title': str(_('Changelog for {name}')).format(name=settings.TX_APP_NAME),
            'content': f'<div class="changelog-wrapper">{content}</div>',
            'submit_label': _("Don't show again") if self.request.GET.get('show') else False,
        }

    def handle_dialog(self):
        """Save shown changelog version"""
        self.request.user.set_attribute('trionyx_last_shown_version', utils.get_app_version())
        return {
            'close': True,
        }


# =============================================================================
# Sidebar
# =============================================================================
class SidebarJsend(JsendView):
    """Model sidebar view"""

    def handle_request(self, request, app, model, pk, code):
        """Return given sidebar"""
        from trionyx.views import sidebars
        config = models_config.get_config(f'{app}.{model}')
        obj = config.model.objects.get(id=pk)
        return sidebars.get_sidebar(config.model, code)(request, obj)


# =============================================================================
# Dashboard
# =============================================================================
class DashboardView(TemplateView):
    """Dashboard view"""

    template_name = 'trionyx/core/dashboard.html'

    def get_context_data(self, *args, **kwargs):
        """Get dashboard context data"""
        context = super().get_context_data(*args, ** kwargs)
        dashboard = self.request.user.get_attribute('tx_dashboard', [])
        if not dashboard:
            for widget in (
                settings.TX_DEFAULT_DASHBOARD()
                if callable(settings.TX_DEFAULT_DASHBOARD)
                else settings.TX_DEFAULT_DASHBOARD
            ):
                widget = widget.copy()
                widget['i'] = utils.random_string(16)
                dashboard.append(widget)

            self.request.user.set_attribute('tx_dashboard', dashboard)

        context.update({
            'widget_templates': [widget().template for index, widget in widgets.items()],
            'widgets': [{
                'code': widget.code,
                'name': str(widget.name),
                'description': str(widget.description),
                'image': static(widget().image),
                'config_fields': widget().config_fields,
                'default_w': widget.default_width,
                'default_h': widget.default_height,
            } for index, widget in widgets.items()],
            'dashboard': dashboard,
        })
        return context


class WidgetDataJsendView(JsendView):
    """Jsend view to get widget data"""

    def handle_request(self, request):
        """Get widget data"""
        data = json.loads(request.body.decode('utf-8'))

        if data['code'] not in widgets:
            raise Exception('Widget does not exists')

        widget = widgets.get(data['code'])()

        return widget.get_data(request, data.get('config', {}))


class SaveDashboardJsendView(JsendView):
    """Jsend view to save user dashboard"""

    def handle_request(self, request):
        """Handle dashboard save"""
        dashboard = json.loads(request.body.decode('utf-8'))

        if not isinstance(dashboard, list):
            raise Exception('Expect dashboard to be a list')

        request.user.set_attribute('tx_dashboard', dashboard)

        return 'saved'


class WidgetConfigDialog(DialogView):
    """Widget config dialog view"""

    def display_dialog(self):
        """Handle widget config"""
        code = self.kwargs.get('code')
        if code not in widgets:
            return {
                'title': _('Widget does not exists')
            }

        if '__post__' in self.request.POST:
            post = self.request.POST
            initial = None
        else:
            post = None
            initial = self.request.POST

        widget = widgets.get(code)
        form = widget.config_form_class(post, initial=initial) if widget.config_form_class else None

        config = None
        if form:
            if not hasattr(form, "helper"):
                form.helper = FormHelper(form)
            form.helper.form_tag = False

            if form.is_valid():
                config = form.cleaned_data
                config['title'] = self.request.POST.get('title')
                config['refresh'] = self.request.POST.get('refresh')
            elif '__post__' in self.request.POST:
                logger.error(json.dumps(form.errors))
        elif '__post__' in self.request.POST:
            config = {
                'title': self.request.POST.get('title'),
                'refresh': self.request.POST.get('refresh')
            }

        return {
            'title': _('Widget config'),
            'content': self.render_to_string('trionyx/dialog/widget_config.html', {
                'form': form,
                'title': self.request.POST.get('title'),
                'refresh': self.request.POST.get('refresh'),
            }),
            'submit_label': _('Save'),
            'config': config
        }

    def handle_dialog(self):
        """Handle widget config"""
        return self.display_dialog()


# Mass actions
class MassDeleteDialog(DialogView):
    """Mass delete dialog"""

    permission_type = 'delete'

    def has_permission(self, request):
        """Check permission"""
        config = self.get_model_config()

        if config.disable_delete:
            return False

        return request.user.has_perm('{app_label}.{type}_{model_name}'.format(
            app_label=config.app_label,
            type=self.permission_type,
            model_name=config.model_name,
        ).lower())

    def display_dialog(self, *args, **kwargs):
        """Handle mass delete"""
        query = self.get_model_class().objects.get_queryset()

        if self.request.POST.get('all', '0') == '1':
            query = filter_queryset_with_user_filters(query, json.loads(self.request.POST.get('filters', '[]')))
        else:
            query = query.filter(id__in=[int(id) for id in filter(None, self.request.POST.get('ids', '').split(','))])

        count = query.count()
        if '__post__' in self.request.POST:
            try:
                query.delete()
            except Exception:
                return {
                    'title': _('Something went wrong on deleting the items'),
                }
            return {
                'deleted': True,
            }
        else:
            return {
                'title': _('Deleting {count} {model_name} items').format(
                    count=count,
                    model_name=self.get_model_config().get_verbose_name(False)
                ),
                'content': self.render_to_string('trionyx/dialog/mass_delete.html', {
                    'count': count,
                    'model_name': self.get_model_config().get_verbose_name(False),
                    'all': self.request.POST.get('all', '0'),
                    'filters': self.request.POST.get('filters', '[]'),
                    'ids': self.request.POST.get('ids', ''),
                }),
                'submit_label': _('Delete') if count else None,
            }

    def handle_dialog(self, *args, **kwargs):
        """Handle mass delete"""
        return self.display_dialog()


class MassUpdateView(ModelPermissionMixin, TemplateView, ModelClassMixin):
    """Mass update view"""

    template_name = 'trionyx/core/model_mass_update.html'
    permission_type = 'change'

    def get(self, *args, **kwargs):
        """Render mass update form"""
        context = self.get_context_data(**kwargs)
        all = self.request.GET.get('all', '0')
        ids = self.request.GET.get('ids', '')
        filters = self.request.GET.get('filters', '[]')
        query = self.get_queryset(all, ids, filters)

        if not query:
            messages.error(self.request, _('You must make a selection'))
            return HttpResponseRedirect(reverse('trionyx:model-list', kwargs=self.kwargs))

        context.update({
            'model_name': self.get_model_config().get_verbose_name(),
            'model_name_plural': self.get_model_config().get_verbose_name_plural(),
            'cancel_url': reverse('trionyx:model-list', kwargs=self.kwargs),
            'all': all,
            'ids': ids,
            'filters': filters,
            'count': query.count(),
            'form': self.get_form(),
            'checked_fields': [],
        })
        return self.render_to_response(context)

    def post(self, *args, **kwargs):
        """Handle mass update form"""
        all = self.request.POST.get('trionyx_all', '0')
        ids = self.request.POST.get('trionyx_ids', '')
        filters = self.request.POST.get('trionyx_filters', '[]')
        query = self.get_queryset(all, ids, filters)

        if not query:
            messages.error(self.request, _('You must make a selection'))
            return HttpResponseRedirect(reverse('trionyx:model-list', kwargs=self.kwargs))

        form = self.get_form(self.request.POST)

        if not form.is_valid():
            context = self.get_context_data(**kwargs)
            context.update({
                'model_name': self.get_model_config().get_verbose_name(),
                'model_name_plural': self.get_model_config().get_verbose_name_plural(),
                'cancel_url': reverse('trionyx:model-list', kwargs=self.kwargs),
                'all': all,
                'ids': ids,
                'filters': filters,
                'count': query.count(),
                'form': form,
                'checked_fields': form.checked_fields,
            })
            return self.render_to_response(context)

        # Get data from form
        data = {}
        for field in form.fields:
            if self.request.POST.get('change_{}'.format(field), False):
                data[field] = form.cleaned_data[field]

        # Start update task
        MassUpdateTask().delay(
            all=all,
            ids=ids,
            filters=filters,
            data=data,

            task_description=_('Mass update {count} {model_name}').format(
                count=query.count(),
                model_name=self.get_model_config().get_verbose_name_plural(),
            ),
            task_model=self.get_model_class(),
        )

        # Add message
        messages.success(self.request, _('Successfully started task for updating {count} {model_name}').format(
            count=query.count(),
            model_name=self.get_model_config().get_verbose_name_plural(),
        ))

        # return to list view
        return HttpResponseRedirect(reverse('trionyx:model-list', kwargs=self.kwargs))

    def get_queryset(self, all, ids, filters):
        """Get queryset"""
        query = self.get_model_class().objects.get_queryset()

        if all == '1':
            return filter_queryset_with_user_filters(query, json.loads(filters))
        else:
            return query.filter(id__in=[int(id) for id in filter(None, ids.split(','))])

    def get_form(self, data=None):
        """Get form class"""
        model_config = self.get_model_config()
        fields = []

        model_fields = [f.name for f in model_config.get_fields()]
        forms = form_register.get_all_forms(self.get_model_class())
        if forms:  # Only use fields that ara also in forms
            for Form in forms:
                fields.extend([name for name in Form().base_fields if name in model_fields])
                fields.extend([name for name in Form().declared_fields if name in model_fields])
        else:
            fields = [f.name for f in model_config.get_fields()]

        FormClass = modelform_factory(self.get_model_class(), fields=sorted(list(set(fields))))
        form = FormClass(data)

        # Disable fields that are not checked for change
        form.checked_fields = []
        for field in form.fields:
            form.fields[field].required = False
            if not (data and data.get('change_{}'.format(field), False)):
                form.fields[field].widget.attrs['readonly'] = True
                form.fields[field].widget.attrs['disabled'] = True
            else:
                form.checked_fields.append(field)

        return form



"""
trionyx.views.dialogs
~~~~~~~~~~~~~~~~~~~~~

:copyright: 2018 by Maikel Martens
:license: GPLv3
"""
import logging
import json
from typing import Optional, Dict, Any, Type

from django.views.generic import View
from django.http import JsonResponse
from django.http.request import HttpRequest
from django.http import Http404
from django.template.loader import render_to_string
from django.db.models import Model
from django.forms import ModelForm
from django.utils.translation import ugettext_lazy as _

from trionyx.views.mixins import ModelClassMixin
from trionyx.forms.helper import FormHelper

logger = logging.getLogger(__name__)


class DialogView(View, ModelClassMixin):
    """
    Dialog view used for showing dialog popup with ajax.

    **Example**

    .. code-block:: python

        # apps/your_app/views.py

        class HelloWorldDialogView(DialogView):

            def display_dialog(self):
                return {
                    "title": "Hello world",
                    "content": self.render_to_string('your_app/dialog/hello_world.html'),
                }

    **Calling in frontend**

    .. code-block:: django

        <button class="btn btn-primary btn-block"
            onClick="openDialog('{% url 'hello:world' %}', {/* dialog options */})">
            Hello World
        </button>
    """

    permission_type: str = 'view'
    """Permission type used for model when no permission is set"""

    permission: Optional[str] = None
    """Permission of dialog, when not set and model is set model view permission is used"""

    model_permission: str = 'read'
    """When no permission is set will use this for model permission"""

    model: Optional[Type[Model]] = None
    """
    Model that will be auto loaded, when set the url mus contain a `pk` parameter.
    Model can also be set with the url with the `app` and `model` parameter.

    Loaded model object is available as self.object and self.<object name lowercase>
    """

    def get(self, request: HttpRequest, *args, **kwargs) -> JsonResponse:
        """Handle get request"""
        try:
            self.load_object(kwargs)
        except Exception as e:
            return self.render_te_response({
                'title': str(e),
            })

        if not self.has_permission(request):
            return self.render_te_response({
                'title': _('No access'),
            })
        return self.render_te_response(self.display_dialog())

    def post(self, request: HttpRequest, *args, **kwargs) -> JsonResponse:
        """Handle post request"""
        try:
            self.load_object(kwargs)
        except Exception as e:
            return self.render_te_response({
                'title': str(e),
            })

        if not self.has_permission(request):
            return self.render_te_response({
                'title': _('No access'),
            })
        return self.render_te_response(self.handle_dialog())

    def load_object(self, kwargs):
        """Load object and model config and remove pk from kwargs"""
        self.object = None
        self.config = None
        try:
            self.model = self.get_model_class()
        except Http404 as e:
            if self.model or ('app' in self.kwargs and 'model' in self.kwargs):
                raise e
            return kwargs

        if self.model and kwargs.get('pk', False):
            try:
                self.object = self.model.objects.get(pk=kwargs.get('pk'))
            except Exception:
                raise Exception(_("Could not load {model}").format(model=self.model.__name__.lower()))
            setattr(self, self.model.__name__.lower(), self.object)

        return kwargs

    def has_permission(self, request: HttpRequest) -> bool:
        """Check if user has permission"""
        if not self.object and not self.permission and self.permission_type != 'add':
            return True

        if not self.permission:
            return self.get_model_config().has_permission(
                self.permission_type,
                self.object,
                request.user
            )

        return request.user.has_perm(self.permission)

    def render_to_string(self, template_file: str, context: Dict[str, Any]) -> str:
        """Render given template to string and add object to context"""
        context = context if context else {}
        if self.object:
            context['object'] = self.object
            context[self.object.__class__.__name__.lower()] = self.object
        return render_to_string(template_file, context, self.request)

    def display_dialog(self) -> Dict[str, Any]:
        """
        Override this function to display a dialog popup. Url params are given as function params.

        The function must return a dict that can contain the following data:

        - **title**: Title of the dialog
        - **content**: Html content to display in dialog, shortcut function :class:`trionyx.trionyx.views.core.DialogView.render_to_string`
        - **url (optional)**: Post url for dialog form, must be a link to a DialogView.
        - **submit_label (optional)**: Label of the submit button, when empty submit button is not shown.
        - **redirect_url (optional)**: Redirect page to given url.
        - **close (optional)**: Close dialog.
        """
        raise NotImplementedError

    def handle_dialog(self) -> Dict[str, Any]:
        """
        Override this function to handle and display popup. Url params are given as function params.
        This function must return the same dict structure as :class:`trionyx.trionyx.views.core.DialogView.display_dialog`

        Post data can be retrieved with *self.request.POST*
        """
        raise NotImplementedError

    def render_te_response(self, data: Dict[str, Any]) -> JsonResponse:
        """Render data to JsonResponse"""
        if 'submit_label' in data and 'url' not in data:
            data['url'] = self.request.get_full_path()

        return JsonResponse(data)


class LayoutDialog(DialogView):
    """Render layout in dialog"""

    def display_dialog(self) -> Dict[str, Any]:
        """Render layout for object"""
        from trionyx.views import layouts
        try:
            content = layouts.get_layout(self.kwargs.get('code'), self.object).render(self.request)
        except Exception:
            content = _('Layout does not exists')

        return {
            'title': str(self.object),
            'content': content,
        }


class UpdateDialog(DialogView):
    """Update dialog view for updating a model"""

    permission_type = 'change'

    template = 'trionyx/dialog/model_form.html'
    """Template for dialog content"""

    title = _("Update {model_name}: {object}")
    """Dialog title model_name and object, variable are given"""

    submit_label = _('save')
    """Dialog submit label value"""

    success_message = _('{model_name} ({object}) is successfully updated')
    """Success message on successfully form saved"""

    def get_form_class(self) -> Type[ModelForm]:
        """Get form class for dialog, default will get form from model config"""
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_edit_form(self.get_model_class())

    def display_dialog(self, form: Optional[ModelForm] = None, success_message: Optional[str] = None) -> Dict[str, Any]:
        """Display form and success message when set"""
        initial = {
            **{key: value for key, value in self.request.GET.items()},
        }

        if not form:
            form = self.get_form_class()(initial=initial, instance=self.object)

        helper = getattr(form, 'helper', FormHelper(form))
        helper.form_tag = False
        setattr(form, 'helper', helper)

        return {
            'title': form.get_title() if hasattr(form, 'get_title') else self.title.format(  # type: ignore
                model_name=self.get_model_config().model_name,
                object=str(self.object) if self.object else '',
            ),
            'content': self.render_to_string(self.template, {
                'form': form,
                'success_message': success_message,
            }),
            'submit_label': form.get_submit_label() if hasattr(form, 'get_submit_label') else self.submit_label,  # type: ignore
            'success': bool(success_message),
        }

    def handle_dialog(self) -> Dict[str, Any]:
        """Handle form and save and set success message on valid form"""
        form = self.get_form_class()(data=self.request.POST, files=self.request.FILES, instance=self.object)

        success_message = None
        if form.is_valid():
            obj = form.save()
            success_message = self.success_message.format(
                model_name=self.get_model_config().get_verbose_name(),
                object=str(obj),
            )
        else:
            logger.debug(json.dumps(form.errors))

        return self.display_dialog(form=form, success_message=success_message)


class CreateDialog(UpdateDialog):
    """Dialog view for creating a model"""

    permission_type = 'add'

    title = _("Create {model_name}")
    submit_label = _('create')
    success_message = _('{model_name} ({object}) is successfully created')

    def get_form_class(self) -> Type[ModelForm]:
        """Get create form class"""
        from trionyx.forms import form_register
        if self.kwargs.get('code'):
            return form_register.get_form(self.get_model_class(), self.kwargs.get('code'))
        return form_register.get_create_form(self.get_model_class())


class DeleteDialog(DialogView):
    """Delete model dialog"""

    permission_type = 'delete'

    template = 'trionyx/dialog/model_delete.html'
    """Template for dialog content"""

    title = _("Delete {model_name}: {object}")
    """Dialog title model_name and object, variable are given"""

    submit_label = _('delete')
    """Dialog submit label value"""

    success_message = _('{model_name} ({object}) is successfully deleted')
    """Success message on successfully deleted"""

    def display_dialog(self) -> Dict[str, Any]:
        """Display delete confirmation"""
        return {
            'title': self.title.format(
                model_name=self.get_model_config().model_name,
                object=str(self.object),
            ),
            'content': self.render_to_string(self.template, {
                'model_name': self.get_model_config().get_verbose_name(),
                'object': self.object,
            }),
            'submit_label': self.submit_label,
        }

    def handle_dialog(self) -> Dict[str, Any]:
        """Handle delete model"""
        object_name = str(self.object)
        response: Dict[str, Any] = {
            'title': self.title.format(
                model_name=self.get_model_config().model_name,
                object=object_name,
            ),
        }

        try:
            self.object.delete()
            response.update({
                'content': self.success_message.format(
                    model_name=self.get_model_config().get_verbose_name(),
                    object=object_name,
                ),
                'success': True,
            })
        except Exception:
            response.update({
                'content': """<p class="alert alert-danger">{}</p>""".format(
                    _('Something went wrong on deleting {model_name}').format(
                        model_name=self.get_model_config().get_verbose_name())
                ),
                'success': False,
            })

        return response



from django.test import TestCase
from django.utils import timezone
from django.conf import settings

from trionyx.trionyx.models import Task, User
from trionyx.trionyx.tasks import cleanup_unexpectedly_stopped_tasks, MassUpdateTask


class TasksTest(TestCase):
    def setUp(self):
        self.user1 = User.objects.create(
            email='info@trionyx.com'
        )
        self.user2 = User.objects.create(
            email='test@test.com'
        )

    def test_mass_update(self):
        Task.objects.create(
            celery_task_id='running-in-time',
            status=Task.RUNNING,
            started_at=timezone.now()
        )
        failed_task = Task.objects.create(
            celery_task_id='running-not-in-time',
            status=Task.RUNNING,
            started_at=timezone.now() - timezone.timedelta(seconds=settings.CELERY_TASK_TIME_LIMIT * 2)
        )

        cleanup_unexpectedly_stopped_tasks()

        failed_task.refresh_from_db()
        self.assertEqual(failed_task.status, Task.FAILED)

    def test_mass_update_selected(self):
        MassUpdateTask().delay(
            all='0',
            ids=str(self.user1.id),
            filters='[]',
            data={
                'first_name': 'new name'
            },

            task_model=User,
        )

        self.user1.refresh_from_db()
        self.user2.refresh_from_db()
        self.assertEqual(self.user1.first_name, 'new name')
        self.assertEqual(self.user2.first_name, '')

    def test_mass_update_filter_all(self):
        MassUpdateTask().delay(
            all='1',
            ids='',
            filters='[{"field": "email", "operator": "==", "value": "test@test.com"}]',
            data={
                'first_name': 'new name'
            },

            task_model=User,
        )

        self.user1.refresh_from_db()
        self.user2.refresh_from_db()
        self.assertEqual(self.user1.first_name, '')
        self.assertEqual(self.user2.first_name, 'new name')

    def test_mass_update_all(self):
        MassUpdateTask().delay(
            all='1',
            ids='',
            filters='[]',
            data={
                'first_name': 'new name'
            },

            task_model=User,
        )

        self.user1.refresh_from_db()
        self.user2.refresh_from_db()
        self.assertEqual(self.user1.first_name, 'new name')
        self.assertEqual(self.user2.first_name, 'new name')



"""
trionyx.signals
~~~~~~~~~~~~~~~

:copyright: 2019 by Maikel Martens
:license: GPLv3
"""
from django.dispatch import Signal

# NOTE: can_view signal wont work for list view or queryset
# NOTE: can_add, can_change, can_delete only works for the interface
# NOTE: can_change, can_delete wont work for mass update/delete
can_view = Signal(providing_args=["instance"])
can_add = Signal(providing_args=["instance"])
can_change = Signal(providing_args=["instance"])
can_delete = Signal(providing_args=["instance"])



from django.dispatch import receiver
from trionyx.signals import can_change

from .models import Post


@receiver(can_change, sender=Post)
def send_pickup_time(sender, instance, **kwargs):
    return instance.id % 2 == 0


"""Reusable app settings"""
import os

from trionyx.settings import *

INTERNAL_IPS = ('127.0.0.1','::1', '0.0.0.0')

INSTALLED_APPS += [
    # Development apps
    'django_extensions',
    'debug_toolbar',
]

MIDDLEWARE = ['debug_toolbar.middleware.DebugToolbarMiddleware'] + list(MIDDLEWARE)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

LOCALE_PATHS = (
    os.path.join(BASE_DIR, '[[name]]', 'locale'),
)

EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
EMAIL_HOST = 'localhost'
EMAIL_PORT = 1025
EMAIL_HOST_USER = ''
EMAIL_HOST_PASSWORD = ''
EMAIL_USE_TLS = False

STATIC_ROOT = os.path.join(BASE_DIR, 'static')
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

COMPRESS_ENABLED = False

# Database
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'development.sqlite3'),
    }
}

# Celery beat schedule
from trionyx.utils import create_celerybeat_schedule
CELERY_BEAT_SCHEDULE = create_celerybeat_schedule(INSTALLED_APPS)

LOGGING = {
    'version': 1,
    'formatters': {
        'color_console': {
            '()': 'colorlog.ColoredFormatter',
            'format': '%(log_color)s%(levelname)-8s [%(name)s:%(lineno)s]%(reset)s %(blue)s %(message)s',
            'datefmt': "%d/%b/%Y %H:%M:%S",
            'log_colors': {
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red',
            },
        },
    },
    'filters': {
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'filters': ['require_debug_true'],
            'class': 'logging.StreamHandler',
            'formatter': 'color_console',
        }
    },
    'loggers': {
        'apps': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        'trionyx': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        'django_jsend': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        # 'django.db.backends': {
        #     'level': 'DEBUG',
        #     'handlers': ['console'],
        # },
        'werkzeug': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
    }
}